"""Evaluator cython extension type to for fast computation of correlation vectors."""

__author__ = "Luis Barroso-Luque"


import cython
import numpy as np
from cython.parallel import prange

cimport numpy as np
from libc.math cimport fabs

from smol.utils.cluster.container cimport (
    FloatArray1DContainer,
    IntArray2DContainer,
    OrbitContainer,
)
from smol.utils.cluster.struct cimport FloatArray1D, IntArray2D, OrbitC


@cython.final
cdef class ClusterSpaceEvaluator(OrbitContainer):
    """ClusterSpaceEvaluator is used to compute correlation and interaction vectors.

    This extension type should not be used directly. Instead, use the
    ClusterSubspace class to create a cluster subspace instance and compute correlation
    vectors using the ClusterSubspace.corr_from_structure method.
    """

    def __cinit__(
            self,
            tuple orbit_data,
            int num_orbits,  # Fix cross-platform dtype issue after numpy>=2.0.
            int num_corr_functions,
            int num_threads = 1,
            double offset = 0.0,
            tuple cluster_interaction_tensors = None
    ):
        """Initialize ClusterSpaceEvaluator extension type.

        Args:
            orbit_data (tuple):
                Tuple of tuples with orbit data. Each tuple contains the following data:
                    - orbit bit id
                    - orbit tensor indices
                    - orbit correlation tensors
            num_orbits (int):
                Number of orbits.
            num_corr_functions (int):
                Number of correlation functions.
            offset (double):
                interaction value for the constant term (i.e. the grand mean).
            cluster_interaction_tensors (tuple):
                Tuple of ndarrays cluster interaction tensors.
        """
        self.num_orbits = num_orbits
        self.num_corr = num_corr_functions
        self.num_threads = num_threads
        self.offset = offset

        if cluster_interaction_tensors is None:
            cluster_interaction_tensors = tuple(data[2].sum(axis=0) for data in orbit_data)

        self.cluster_interactions = FloatArray1DContainer(cluster_interaction_tensors)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return (
            ClusterSpaceEvaluator,
            (
                self._orbit_data,
                self.num_orbits,
                self.num_corr,
                self.num_threads,
                self.offset,
                self.cluster_interactions._arrays,
            ),
        )

    cpdef public void reset_data(
            self,
            tuple orbit_data,
            int num_orbits,
            int num_corr_functions,
    ):
        """Reset data of ClusterSpaceEvaluator extension type.

        Args:
            orbit_data (tuple):
                Tuple of tuples with orbit data. Each tuple contains the following data:
                    - orbit bit id
                    - orbit tensor indices
                    - orbit correlation tensors
            num_orbits (int):
                Number of orbits.
            num_corr_functions (int):
                Number of correlation functions.
        """
        self.num_orbits = num_orbits
        self.num_corr = num_corr_functions
        self.set_orbits(orbit_data)

    cpdef public void set_cluster_interactions(
            self, tuple cluster_interaction_tensors, double offset
    ):
        """Sets the cluster interaction tensors.

        Args:
            cluster_interaction_tensors (tuple):
                Tuple of ndarrays cluster interaction tensors.
            offset (double):
                interaction value for the constant term (i.e. the grand mean).
        """
        if len(cluster_interaction_tensors) != self.size:
            raise ValueError(
                "Number of cluster interaction tensors must be equal to the number of orbits."
            )
        self.cluster_interactions.set_arrays(cluster_interaction_tensors)
        self.offset = offset

    cpdef np.ndarray[np.float64_t, ndim=1] correlations_from_occupancy(
            self,
            const np.int32_t[::1] occu,
            IntArray2DContainer cluster_indices,
    ):
        """Computes the correlation vector for a given encoded occupancy string.

        Args:
            occu (ndarray):
                encoded occupancy vector
            cluster_indices (IntArray1DContainer):
                Container with pointers to arrays with indices of sites of all clusters
                in each orbit given as a container of arrays.

        Returns: array
            correlation vector for given occupancy
        """
        cdef int i, j, k, n, I, J, K, N, index, bit_id
        cdef double p
        cdef IntArray2D indices  # flattened tensor indices
        cdef OrbitC orbit

        out = np.zeros(self.num_corr)
        cdef np.float64_t[::1] o_view = out
        o_view[0] = 1  # empty cluster

        # loop thru orbits
        for n in prange(self.size, nogil=True, schedule="guided", num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            bit_id = orbit.bit_id
            K = orbit.correlation_tensors.size_r  # index of bit combos
            J = indices.size_r # cluster index
            I = indices.size_c # index within cluster
            N = orbit.correlation_tensors.size_c # size of single flattened tensor

            for k in range(K):  # loop thru bit combos
                p = 0
                for j in range(J):  # loop thru clusters
                    index = 0
                    for i in range(I):  # loop thru sites in cluster
                        index = index + orbit.tensor_indices.data[i] * occu[indices.data[j * I + i]]
                    # sum contribution of correlation of cluster k with occupancy at "index"
                    p = p + orbit.correlation_tensors.data[k * N + index]
                o_view[bit_id] = p / J
                bit_id = bit_id + 1

        return out

    cpdef np.ndarray[np.float64_t, ndim=1] interactions_from_occupancy(
            self,
            const np.int32_t[::1] occu,
            IntArray2DContainer cluster_indices,
    ):
        """Computes the cluster interaction vector for a given encoded occupancy string.
        Args:
            occu (ndarray):
                encoded occupancy vector
            cluster_indices (IntArray1DContainer):
                Container with pointers to arrays with indices of sites of all clusters
                in each orbit given as a container of arrays.
        Returns: array
            cluster interaction vector for given occupancy
        """
        cdef int n, i, j, I, J, index
        cdef double p
        cdef IntArray2D indices
        cdef OrbitC orbit
        cdef FloatArray1D interaction_tensor

        out = np.zeros(self.num_orbits)
        cdef np.float64_t[::1] o_view = out
        o_view[0] = self.offset  # empty cluster

        for n in prange(self.size, nogil=True, schedule="guided", num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            interaction_tensor = self.cluster_interactions.data[n]
            J = indices.size_r # cluster index
            I = indices.size_c  # index within cluster
            p = 0
            for j in range(J):
                index = 0
                for i in range(I):
                    index = index + orbit.tensor_indices.data[i] * occu[indices.data[j * I + i]]
                p = p + interaction_tensor.data[index]
            o_view[orbit.id] = p / J

        return out

    cpdef np.ndarray[np.float64_t, ndim=1] delta_correlations_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] cluster_ratio,
            IntArray2DContainer cluster_indices,
    ):
        """Computes the correlation difference between two occupancy vectors.

        Args:
            occu_f (ndarray):
                encoded occupancy vector with flip
            occu_i (ndarray):
                encoded occupancy vector without flip
            cluster_ratio (ndarray):
                ratio of number of clusters in each entry of cluster_indices to the total
                number of the clusters in a structure for the corresponding cluster.
            cluster_indices (IntArray2DContainer):
                Container with pointers to arrays with indices of sites of all clusters
                in each orbit given as a container of arrays.

        Returns: ndarray
            correlation vector difference
        """
        cdef int i, j, k, n, I, J, K, N, ind_i, ind_f, bit_id
        cdef double p
        cdef IntArray2D indices  # flattened tensor indices
        cdef OrbitC orbit

        out = np.zeros(self.num_corr)
        cdef np.float64_t[::1] o_view = out

        # loop thru orbits
        for n in prange(self.size, nogil=True, num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            bit_id = orbit.bit_id
            K = orbit.correlation_tensors.size_r  # index of bit combos
            J = indices.size_r # cluster index
            I = indices.size_c # index within cluster
            N = orbit.correlation_tensors.size_c # size of single flattened tensor

            for k in range(K):  # loop thru bit combos
                p = 0
                for j in range(J):  # loop thru clusters
                    ind_i, ind_f = 0, 0
                    for i in range(I):  # loop thru sites in cluster
                        ind_i = ind_i + orbit.tensor_indices.data[i] * occu_i[indices.data[j * I + i]]
                        ind_f = ind_f + orbit.tensor_indices.data[i] * occu_f[indices.data[j * I + i]]
                    # sum contribution of correlation of cluster k with occupancy at "index"
                    p = p + (orbit.correlation_tensors.data[k * N + ind_f] - orbit.correlation_tensors.data[k * N + ind_i])
                o_view[bit_id] = p / cluster_ratio[n] / J
                bit_id = bit_id + 1

        return out

    cpdef np.ndarray[np.float64_t, ndim=1] delta_interactions_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] cluster_ratio,
            IntArray2DContainer cluster_indices,

    ):
        """Computes the cluster interaction vector difference between two occupancy
        strings.

        Args:
            occu_f (ndarray):
                encoded occupancy vector with flip
            occu_i (ndarray):
                encoded occupancy vector without flip
            cluster_ratio (ndarray):
                ratio of number of clusters in each entry of cluster_indices to the total
                number of the clusters in a structure for the corresponding cluster.
            cluster_indices (IntArray2DContainer):
                Container with pointers to arrays with indices of sites of all clusters
                in each orbit given as a container of arrays.

        Returns: ndarray
            cluster interaction vector difference
        """
        cdef int i, j, n, I, J, ind_i, ind_f
        cdef double p
        cdef IntArray2D indices
        cdef OrbitC orbit
        cdef FloatArray1D interaction_tensor

        out = np.zeros(self.num_orbits)
        cdef np.float64_t[::1] o_view = out

        for n in prange(self.size, nogil=True, num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            interaction_tensor = self.cluster_interactions.data[n]
            J = indices.size_r # cluster index
            I = indices.size_c # index within cluster
            p = 0
            for j in range(J):
                ind_i, ind_f = 0, 0
                for i in range(I):
                    ind_i = ind_i + orbit.tensor_indices.data[i] * occu_i[indices.data[j * I + i]]
                    ind_f = ind_f + orbit.tensor_indices.data[i] * occu_f[indices.data[j * I + i]]
                p = p + (interaction_tensor.data[ind_f] - interaction_tensor.data[ind_i])
            o_view[orbit.id] = p / cluster_ratio[n] / J

        return out

    cpdef np.ndarray[np.float64_t, ndim=1] corr_distances_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] ref_corr_vector,
            IntArray2DContainer cluster_indices
    ):
        """Computes the absolute distance of two correlation vectors separated by a single
        flip and a given correlation vector.

        Unfortunately this scales just as bad as computing the full correlation vector.

        Args:
            occu_f (ndarray):
                encoded occupancy array with flip
            occu_i (ndarray):
                encoded occupancy array without flip
            ref_corr_vector (ndarray):
                reference correlation vector
            cluster_indices (IntArray2DContainer):
                Container with pointers to arrays with indices of sites of all clusters
                in each orbit given as a container of arrays.

        Returns:
            ndarray: 2D with correlation vector distances from reference for each of occu_i
            and occu_f
        """
        cdef int i, j, k, n, I, J, K, N, bit_id, ind_i, ind_f
        cdef double p_i, p_f
        cdef IntArray2D indices  # flattened tensor indices
        cdef OrbitC orbit

        out = np.zeros((2, self.num_corr))
        cdef np.float64_t[:, ::1] o_view = out
        o_view[:, 0] = 0

        # loop thru orbits
        for n in prange(self.size, nogil=True, schedule="guided", num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            bit_id = orbit.bit_id
            K = orbit.correlation_tensors.size_r  # index of bit combos
            J = indices.size_r # cluster index
            I = indices.size_c # index within cluster
            N = orbit.correlation_tensors.size_c # size of single flattened tensor

            for k in range(K):  # loop thru bit combos
                p_f, p_i = 0, 0
                for j in range(J):  # loop thru clusters
                    ind_f, ind_i = 0, 0
                    for i in range(I):  # loop thru sites in cluster
                        ind_f = ind_f + orbit.tensor_indices.data[i] * occu_f[indices.data[j * I + i]]
                        ind_i = ind_i + orbit.tensor_indices.data[i] * occu_i[indices.data[j * I + i]]
                    # sum contribution of correlation of cluster k with occupancy at "ind"
                    p_f = p_f + orbit.correlation_tensors.data[k * N + ind_f]
                    p_i = p_i + orbit.correlation_tensors.data[k * N + ind_i]
                o_view[1, bit_id] = fabs(p_f / J - ref_corr_vector[bit_id])
                o_view[0, bit_id] = fabs(p_i / J - ref_corr_vector[bit_id])
                bit_id = bit_id + 1

        return out

    cpdef np.ndarray[np.float64_t, ndim=1] interaction_distances_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] ref_interaction_vector,
            IntArray2DContainer cluster_indices
    ):
        """Computes the absolute distance of two cluster interaction vectors separated by a
            single flip and a given correlation vector.

            Unfortunately this scales just as bad as computing the full interaction vector.

            Args:
                occu_f (ndarray):
                    encoded occupancy array with flip
                occu_i (ndarray):
                    encoded occupancy array without flip
                ref_interaction_vector (ndarray):
                    reference cluster interaction vector
                cluster_indices (IntArray2DContainer):
                    Container with pointers to arrays with indices of sites of all clusters
                    in each orbit given as a container of arrays.

            Returns:
                ndarray: 2D with cluster interaction vector distances from reference for each of
                occu_i and occu_f
            """
        cdef int n, i, j, I, J, ind_i, ind_f
        cdef double p_i, p_f
        cdef IntArray2D indices
        cdef OrbitC orbit
        cdef FloatArray1D interaction_tensor

        out = np.zeros((2, self.num_orbits))
        cdef np.float64_t[:, ::1] o_view = out
        o_view[:, 0] = 0

        for n in prange(self.size, nogil=True, schedule="guided", num_threads=self.num_threads):
            orbit = self.data[n]
            indices = cluster_indices.data[n]
            interaction_tensor = self.cluster_interactions.data[n]
            J = indices.size_r # cluster index
            I = indices.size_c  # index within cluster
            p_f, p_i = 0, 0
            for j in range(J):
                ind_f, ind_i = 0, 0
                for i in range(I):
                    ind_f = ind_f + orbit.tensor_indices.data[i] * occu_f[indices.data[j * I + i]]
                    ind_i = ind_i + orbit.tensor_indices.data[i] * occu_i[indices.data[j * I + i]]
                p_f = p_f + interaction_tensor.data[ind_f]
                p_i = p_i + interaction_tensor.data[ind_i]
            o_view[1, orbit.id] = fabs(p_f / J - ref_interaction_vector[orbit.id])
            o_view[0, orbit.id] = fabs(p_i / J - ref_interaction_vector[orbit.id])

        return out



import numpy as np
import pytest

from smol.moca.analysis.convergence import (
    check_property_converged,
    determine_discard_number,
)


@pytest.fixture()
def prop_array_1():  # array with an outlier at the end
    array1 = [1, 2] * 100
    array1.append(10)
    return np.array(array1)


@pytest.fixture()
def prop_array_2(rng):
    # linearly decreasing array with some noise, then hold constant with noise
    array2 = []
    for i in np.arange(1, -0.001, -0.05):
        array2.append(i)
        random_samples = rng.normal(loc=i, scale=0.01, size=10)
        array2.extend(random_samples)

    more_random_samples = rng.normal(loc=0, scale=0.01, size=10000)
    array2.extend(more_random_samples)
    array2.append(0.005)
    return np.array(array2)


def test_check_property_converged(prop_array_1, prop_array_2):
    assert not check_property_converged(prop_array_1)
    assert not check_property_converged(prop_array_2[:200])
    assert check_property_converged(prop_array_2[1000:])


def test_determine_discard_number(prop_array_1, prop_array_2):
    assert not determine_discard_number(prop_array_1)
    assert not determine_discard_number(prop_array_2[:200])
    assert determine_discard_number(prop_array_2[1000:]) > 0



import itertools
import re
from copy import deepcopy

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Composition, Species

from smol.cofe.space.domain import Vacancy
from smol.moca.composition.space import CompositionSpace
from smol.moca.occu_utils import get_dim_ids_by_sublattice
from smol.utils.math import NUM_TOL, compute_snf, integerize_vector
from tests.utils import assert_msonable, assert_pickles, assert_table_set_equal


@pytest.fixture(scope="module")
def comp_space(ensemble):
    bits = [s.species for s in ensemble.sublattices]
    sublattice_sizes = np.array([len(s.sites) for s in ensemble.sublattices])
    gcd = np.gcd.reduce(sublattice_sizes)
    sublattice_sizes = sublattice_sizes // gcd
    return CompositionSpace(bits, sublattice_sizes)  # Charge balanced = True, default.


@pytest.fixture(scope="module")
def comp_space_lmtpo():
    li = Species.from_str("Li+")
    mn = Species.from_str("Mn3+")
    ti = Species.from_str("Ti4+")
    o = Species.from_str("O2-")
    p = Species.from_str("P3-")

    bits = [[li, mn, ti], [p, o]]
    sublattice_sizes = [1, 1]
    return CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
        optimize_basis=True,
        table_ergodic=True,
    )


@pytest.fixture(scope="module")
def comp_space_lmtpo2():
    li = Species.from_str("Li+")
    mn = Species.from_str("Mn3+")
    ti = Species.from_str("Ti4+")
    o = Species.from_str("O2-")
    p = Species.from_str("P3-")

    bits = [[li, mn, ti], [p, o]]
    sublattice_sizes = [1, 1]
    return CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
        optimize_basis=True,
        table_ergodic=True,
        other_constraints=[([2, 1, 0, 0, 0], 7 / 6, "eq")],
    )


@pytest.fixture(scope="module")
def comp_space_lmtpo3():
    li = Species.from_str("Li+")
    mn = Species.from_str("Mn3+")
    ti = Species.from_str("Ti4+")
    o = Species.from_str("O2-")
    p = Species.from_str("P3-")

    bits = [[li, mn, ti], [p, o]]
    sublattice_sizes = [1, 1]
    return CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
        optimize_basis=True,
        table_ergodic=True,
        other_constraints=[
            ([0, 1, 0, 0, 0], 5 / 6, "<="),
            ([0, 0, 0, 1, 0], 5 / 6, "<="),
            ([2, 1, 0, 0, 0], 8 / 6, "<="),
            ([0, 1, 0, 0, 0], 1 / 6, ">="),
            ([0, 0, 0, 1, 0], 1 / 6, ">="),
            ([2, 1, 0, 0, 0], 5 / 6, ">="),
        ],
    )


# A test example with extra constraints.
@pytest.fixture(scope="module")
def comp_space_lmntof():
    li = Species.from_str("Li+")
    ni = Species.from_str("Ni2+")
    mn = Species.from_str("Mn3+")
    ti = Species.from_str("Ti4+")
    f = Species.from_str("F-")
    o = Species.from_str("O2-")

    bits = [[li, ni, mn, ti], [o, f]]
    sublattice_sizes = [1, 1]
    other_constraints = [([0, 1, -1, 0, 0, 0], 0, "eq"), ([0, 0, 1, -1, 0, 0], 0, "eq")]
    return CompositionSpace(
        bits,
        sublattice_sizes,
        other_constraints=other_constraints,
        charge_neutral=True,
        optimize_basis=True,
        table_ergodic=True,
    )


# Generic attributes test.
def test_generic_attributes(comp_space):
    dim_ids = comp_space.dim_ids
    assert dim_ids == get_dim_ids_by_sublattice(comp_space.site_spaces)
    species_set = set(itertools.chain(*comp_space.site_spaces))
    vac_count = 0
    for sp in comp_space.species:
        assert sp in species_set
        if isinstance(sp, Vacancy):
            vac_count += 1
    assert vac_count <= 1
    assert species_set & set(comp_space.species) == set(comp_space.species)
    assert sorted(comp_space.species) == comp_space.species

    for sl_id, dim_ids_sl in enumerate(comp_space.species_ids):
        for sp_id, d in enumerate(dim_ids_sl):
            bit = comp_space.site_spaces[sl_id][sp_id]
            sp = comp_space.species[d]
            if not isinstance(sp, Vacancy):
                assert bit == sp
            else:
                assert isinstance(bit, Vacancy)

    A = comp_space._A
    b = comp_space._b
    assert len(A) == 1 + len(comp_space.site_spaces)  # No other constraints.
    assert len(b) == len(A)
    assert b[0] == 0
    assert np.all(A[1:, :] <= 1)
    npt.assert_array_equal(b[1:], comp_space.sublattice_sizes)
    min_sc = comp_space.min_supercell_size

    prim_verts = comp_space.prim_vertices
    assert prim_verts.shape[1] == A.shape[1]
    npt.assert_almost_equal(A @ prim_verts.T - b[:, None], 0, decimal=6)
    assert np.all(np.any(np.isclose(prim_verts, 0, atol=NUM_TOL), axis=-1))

    if comp_space.num_unconstrained_compositions < 10**6:
        scs = [1, min_sc, min_sc * 2] + ([min_sc // 2] if min_sc >= 2 else [])
        for sc in scs:
            if not np.allclose(sc * comp_space._b, np.round(sc * comp_space._b)):
                with pytest.raises(ValueError):
                    grid = comp_space.get_composition_grid(sc, step=1)
            else:
                grid = comp_space.get_composition_grid(sc, step=1)
            ns = grid @ comp_space.basis + comp_space.get_supercell_base_solution(sc)
            npt.assert_array_equal(grid, comp_space._comp_grids[(sc, 1)])
            # x-format
            npt.assert_array_equal(A @ ns.T - b[:, None] * sc, 0)
            assert np.all(ns >= 0)

    # Flip table not optimized.
    npt.assert_array_equal(comp_space.flip_table, comp_space.basis)

    reactions = comp_space.flip_reactions
    flips = []
    bits_str = [[str(sp) for sp in sl_sps] for sl_sps in comp_space.site_spaces]
    for r in reactions:
        sps = []
        sl_ids = []
        nums = []
        n_from = np.inf
        tub_id = 0
        for t in r.split():
            if t not in ["->", "+"]:
                if t[-1] == ")":
                    sp_str, sl_id = re.search(r"(.+)\((\d+)\)", t).groups()
                    sps.append(sp_str)
                    sl_ids.append(int(sl_id))
                    tub_id += 1
                else:
                    if tub_id < n_from:
                        nums.append(-int(t))
                    else:
                        nums.append(int(t))
            elif t == "->":
                n_from = tub_id
        flip = np.zeros(comp_space.num_dims, dtype=int)

        for sl_id, sp_str, n in zip(sl_ids, sps, nums):
            sp_id = bits_str[sl_id].index(sp_str)
            ii = comp_space.dim_ids[sl_id][sp_id]
            flip[ii] += n
        flips.append(flip)
    flips = np.array(flips, dtype=int)
    npt.assert_array_equal(flips, comp_space.flip_table)


def test_serialize(comp_space, comp_space_lmtpo3, comp_space_lmntof):
    _ = comp_space.flip_table
    if comp_space.num_unconstrained_compositions < 10**6:
        _ = comp_space.min_supercell_grid
    assert_msonable(comp_space)
    assert_msonable(comp_space_lmtpo3)
    assert_msonable(comp_space_lmntof)

    assert_pickles(comp_space)
    assert_pickles(comp_space_lmtpo3)
    assert_pickles(comp_space_lmntof)

    comp_space_reload = CompositionSpace.from_dict(comp_space.as_dict())
    npt.assert_array_equal(comp_space_reload._flip_table, comp_space._flip_table)
    npt.assert_array_equal(comp_space_reload._vs, comp_space._vs)

    npt.assert_array_equal(comp_space_reload._n0, comp_space._n0)

    npt.assert_array_equal(
        comp_space_reload._min_supercell_size, comp_space._min_supercell_size
    )
    assert set(list(comp_space._comp_grids.keys())) == set(
        list(comp_space_reload._comp_grids.keys())
    )
    for k in comp_space._comp_grids.keys():
        npt.assert_array_equal(
            comp_space._comp_grids[k], comp_space_reload._comp_grids[k]
        )


# Pre-computed data test.
def test_supercell_size(
    comp_space_lmtpo, comp_space_lmtpo2, comp_space_lmtpo3, comp_space_lmntof
):
    assert comp_space_lmtpo.min_supercell_size == 6
    assert comp_space_lmtpo2.min_supercell_size == 12
    assert comp_space_lmtpo3.min_supercell_size == 6
    assert comp_space_lmntof.min_supercell_size == 6


def test_n_comps_unconstrained(comp_space_lmtpo, comp_space_lmntof):
    assert comp_space_lmtpo.num_unconstrained_compositions == 7776
    assert comp_space_lmntof.num_unconstrained_compositions == 46656


def test_basis(
    comp_space_lmtpo, comp_space_lmtpo2, comp_space_lmtpo3, comp_space_lmntof
):
    vs1 = comp_space_lmtpo.basis
    vs2 = comp_space_lmntof.basis
    vs3 = comp_space_lmtpo2.basis
    vs4 = comp_space_lmtpo3.basis
    ts1 = comp_space_lmtpo.flip_table
    ts2 = comp_space_lmntof.flip_table
    ts3 = comp_space_lmtpo2.flip_table
    ts4 = comp_space_lmtpo3.flip_table
    # Pre-computed optimal basis
    std1 = np.array([[0, -1, 1, 1, -1], [-1, 1, 0, 2, -2]], dtype=int)
    std2 = np.array([[-3, 1, 1, 1, 6, -6]], dtype=int)
    std3 = np.array([[-1, 2, -1, 1, -1]], dtype=int)

    assert_table_set_equal(vs1, std1)
    assert_table_set_equal(vs2, std2)
    assert_table_set_equal(ts1, std1)
    assert_table_set_equal(ts2, std2)
    assert_table_set_equal(vs3, std3)
    assert_table_set_equal(vs4, std1)
    assert_table_set_equal(ts3, std3)
    assert_table_set_equal(ts4, std1)


def test_get_supercell_base_solution(comp_space):
    comp_space._n0 = None  # Clear up.
    min_supercell_size = comp_space.min_supercell_size
    n00 = comp_space.get_supercell_base_solution(supercell_size=min_supercell_size * 2)
    _, min_feasible_size = integerize_vector(comp_space._b)
    assert min_supercell_size % min_feasible_size == 0
    scale = min_supercell_size * 2 // min_feasible_size
    n01 = comp_space.get_supercell_base_solution(supercell_size=min_feasible_size)
    npt.assert_array_equal(n01 * scale, n00)


def test_enumerate_grid(comp_space_lmtpo, comp_space_lmtpo2, comp_space_lmtpo3):
    # Convert x-format to n-format
    grid2 = (
        comp_space_lmtpo2.min_supercell_grid @ comp_space_lmtpo2.basis
        + comp_space_lmtpo2.get_supercell_base_solution()
    )
    grid3 = (
        comp_space_lmtpo3.min_supercell_grid @ comp_space_lmtpo3.basis
        + comp_space_lmtpo3.get_supercell_base_solution()
    )
    std2 = np.array(
        [
            [2, 10, 0, 8, 4],
            [3, 8, 1, 7, 5],
            [4, 6, 2, 6, 6],
            [5, 4, 3, 5, 7],
            [6, 2, 4, 4, 8],
            [7, 0, 5, 3, 9],
        ],
        dtype=int,
    )
    std3 = np.array(
        [
            [3, 2, 1, 1, 5],
            [3, 1, 2, 2, 4],
            [2, 4, 0, 2, 4],
            [2, 3, 1, 3, 3],
            [2, 2, 2, 4, 2],
            [1, 5, 0, 4, 2],
            [2, 1, 3, 5, 1],
            [1, 4, 1, 5, 1],
        ],
        dtype=int,
    )
    grid2 = np.array(sorted(grid2.tolist()), dtype=int)
    grid3 = np.array(sorted(grid3.tolist()), dtype=int)
    std2 = np.array(sorted(std2.tolist()), dtype=int)
    std3 = np.array(sorted(std3.tolist()), dtype=int)
    npt.assert_array_equal(grid2, std2)
    npt.assert_array_equal(grid3, std3)

    grid = comp_space_lmtpo.get_composition_grid(
        supercell_size=4
    ) @ comp_space_lmtpo.basis + comp_space_lmtpo.get_supercell_base_solution(
        supercell_size=4
    )
    std = np.array(
        [
            [0, 4, 0, 4, 0],
            [1, 2, 1, 3, 1],
            [2, 0, 2, 2, 2],
            [1, 3, 0, 2, 2],
            [1, 1, 2, 4, 0],
            [2, 1, 1, 1, 3],
            [2, 2, 0, 0, 4],
        ],
        dtype=int,
    )
    grid = np.array(sorted(grid.tolist()), dtype=int)
    std = np.array(sorted(std.tolist()), dtype=int)
    npt.assert_array_equal(grid, std)

    # Scalability guaranteed in comp_space, which means when supercell_size / step
    # is the same value, the enumerated grid will always be original_grid
    # * step.
    min_supercell_size = comp_space_lmtpo.min_supercell_size
    grid = comp_space_lmtpo.get_composition_grid(
        supercell_size=2 * min_supercell_size, step=2
    )
    std = comp_space_lmtpo.min_supercell_grid * 2
    grid = np.array(sorted(grid.tolist()), dtype=int)
    std = np.array(sorted(std.tolist()), dtype=int)
    npt.assert_array_equal(grid, std)

    grid = comp_space_lmtpo.get_composition_grid(supercell_size=8, step=2)
    std = comp_space_lmtpo.get_composition_grid(supercell_size=4, step=1) * 2
    std2 = comp_space_lmtpo.get_composition_grid(supercell_size=4, step=1) * 2
    grid = np.array(sorted(grid.tolist()), dtype=int)
    std = np.array(sorted(std.tolist()), dtype=int)
    std2 = np.array(sorted(std2.tolist()), dtype=int)
    npt.assert_array_equal(grid, std)
    npt.assert_array_equal(grid, std2)

    grid = comp_space_lmtpo2.get_composition_grid(supercell_size=12, step=2)
    std = comp_space_lmtpo2.get_composition_grid(supercell_size=6, step=1) * 2
    std2 = comp_space_lmtpo2.get_composition_grid(supercell_size=6, step=1) * 2
    grid = np.array(sorted(grid.tolist()), dtype=int)
    std = np.array(sorted(std.tolist()), dtype=int)
    std2 = np.array(sorted(std2.tolist()), dtype=int)
    npt.assert_array_equal(grid, std)
    npt.assert_array_equal(grid, std2)

    grid = comp_space_lmtpo3.get_composition_grid(supercell_size=8, step=2)
    std = comp_space_lmtpo3.get_composition_grid(supercell_size=4, step=1) * 2
    std2 = comp_space_lmtpo3.get_composition_grid(supercell_size=4, step=1) * 2
    grid = np.array(sorted(grid.tolist()), dtype=int)
    std = np.array(sorted(std.tolist()), dtype=int)
    std2 = np.array(sorted(std2.tolist()), dtype=int)
    npt.assert_array_equal(grid, std)
    npt.assert_array_equal(grid, std2)
    # Test inequality constraints are all correctly parsed and satisfied.
    ns = [
        comp_space_lmtpo3.translate_format(
            x,
            supercell_size=8,
            from_format="coordinates",
            to_format="counts",
            rounding=True,
        )
        for x in grid
    ]
    ns = np.array(ns, dtype=int)
    # No geq now!
    assert comp_space_lmtpo3._A_leq.shape == (6, ns.shape[1])
    a_leq = comp_space_lmtpo3._A_leq
    b_leq = comp_space_lmtpo3._b_leq

    assert np.all(a_leq @ ns.T / 8 <= b_leq[:, None] + NUM_TOL)

    grid1 = comp_space_lmtpo.get_composition_grid(supercell_size=10, step=2)
    grid2 = comp_space_lmtpo.get_composition_grid(supercell_size=5, step=1) * 2
    grid1 = np.array(sorted(grid1.tolist()), dtype=int)
    grid2 = np.array(sorted(grid2.tolist()), dtype=int)
    npt.assert_array_equal(grid1, grid2)


def test_grid_storage(comp_space_lmtpo):
    comp_space_lmtpo._comp_grids = {}  # Clear up for new enumeration
    assert len(comp_space_lmtpo._comp_grids) == 0
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=10, step=2)
    _ = comp_space_lmtpo.get_composition_grid()
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=24, step=4)
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=16, step=6)
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=8, step=3)
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=12, step=1)
    _ = comp_space_lmtpo.get_composition_grid(supercell_size=24, step=2)
    keyset = {(5, 1), (6, 1), (8, 3), (12, 1), (1, 1)}
    assert keyset == set(list(comp_space_lmtpo._comp_grids.keys()))


def test_convert_formats(comp_space):
    s, m, t = compute_snf(comp_space._A)
    null_space = t[:, : comp_space.num_dims - len(comp_space.basis)].T
    for _ in range(5):
        # Test good cases.
        supercell_size = (
            20 // comp_space.min_supercell_size * comp_space.min_supercell_size
        )
        a = comp_space._A
        b = comp_space._b * supercell_size
        x_std = comp_space.get_centroid_composition(supercell_size)
        # print("basis:", comp_space.basis)
        # print("n0:", comp_space.get_supercell_base_solution(supercell_size))
        # print("x_std:", x_std)
        n = comp_space.basis.T @ x_std + comp_space.get_supercell_base_solution(
            supercell_size
        )
        npt.assert_almost_equal(a @ n - b, 0, decimal=6)
        assert np.all(n >= -NUM_TOL)

        n2 = comp_space.translate_format(
            n, supercell_size, from_format="counts", to_format="counts"
        )
        npt.assert_almost_equal(n2, n, decimal=6)
        x = comp_space.translate_format(
            n, supercell_size, from_format="counts", to_format="coordinates"
        )
        # This x format must be integers.
        npt.assert_almost_equal(x, np.round(x), decimal=6)
        n0 = comp_space.get_supercell_base_solution(supercell_size)
        npt.assert_almost_equal(comp_space.basis.T @ x + n0, n, decimal=6)
        npt.assert_almost_equal(x, x_std, decimal=6)
        npt.assert_almost_equal(
            comp_space.translate_format(
                x, supercell_size, from_format="coordinates", to_format="counts"
            ),
            n,
        )
        c = comp_space.translate_format(
            n, supercell_size, from_format="counts", to_format="compositions"
        )
        assert len(c) == len(comp_space.site_spaces)
        assert all(isinstance(sl_c, Composition) for sl_c in c)
        assert all(0 <= sl_c.num_atoms <= 1 for sl_c in c)
        npt.assert_almost_equal(
            comp_space.translate_format(
                c, supercell_size, from_format="compositions", to_format="counts"
            ),
            n,
            decimal=6,
        )
        nd = comp_space.translate_format(
            n, supercell_size, from_format="counts", to_format="species-counts"
        )
        for sp_id, sp in enumerate(comp_space.species):
            dim_id = []
            for dim_ids, sl_sps in zip(comp_space.dim_ids, comp_space.site_spaces):
                for d, sp2 in zip(dim_ids, sl_sps):
                    if not isinstance(sp, Vacancy):
                        if sp == sp2:
                            dim_id.append(d)
                    elif isinstance(sp2, Vacancy):
                        dim_id.append(d)
            assert np.sum(n[dim_id]) == nd[sp_id]
        # Nondisc format can not be converted to anything else.
        with pytest.raises(ValueError):
            _ = comp_space.translate_format(
                nd, supercell_size, from_format="species-counts", to_format="counts"
            )
        with pytest.raises(ValueError):
            _ = comp_space.translate_format(
                n, supercell_size, from_format="counts", to_format="whatever"
            )
        with pytest.raises(ValueError):
            _ = comp_space.translate_format(
                n, supercell_size, from_format="whatever", to_format="counts"
            )
        # Test bad cases.
        # Test negative error.
        n_bad = n - 30
        with pytest.raises(ValueError):
            _ = comp_space.translate_format(
                n_bad, supercell_size, from_format="counts", to_format="counts"
            )
        # Test constraints violation
        null_vec = np.random.rand(len(null_space)) * 0.01
        dn = null_space.T @ null_vec
        n_bad = n + dn
        if len(null_space) > 0 and not np.allclose(
            np.abs(comp_space._A @ (n / supercell_size) - comp_space._b),
            0,
            atol=NUM_TOL,
        ):
            if np.all(n_bad >= 0):
                # print("bad composition:", n_bad)
                # print("A:", comp_space._A)
                # print("b:", comp_space._b * supercell_size)
                with pytest.raises(ValueError):
                    _ = comp_space.translate_format(
                        n_bad, supercell_size, from_format="counts", to_format="counts"
                    )
        # Test rounding fails
        dx = np.random.rand(len(comp_space.basis)) + 1e-5
        x_bad = x + dx
        n_bad = comp_space.basis.T @ x_bad + n0
        if np.all(n_bad >= 0):
            with pytest.raises(ValueError):
                _ = comp_space.translate_format(
                    n_bad,
                    supercell_size,
                    from_format="counts",
                    to_format="counts",
                    rounding=True,
                )
        # Test un-normalized compositions.
        c_bad = deepcopy(c)
        c0_bad = {k: v + 100 for k, v in c_bad[0].items()}
        c_bad[0] = Composition(c0_bad)
        with pytest.raises(ValueError):
            _ = comp_space.translate_format(
                c_bad, supercell_size, from_format="compositions", to_format="counts"
            )



"""A metadata class to recorded sampling metadata in MCKernels, MCUshers, MCBias classes."""

__author__ = "Luis Barroso-Luque"

from types import SimpleNamespace

from monty.json import MontyDecoder, MSONable, jsanitize


class Metadata(SimpleNamespace, MSONable):
    """A simple namespace to hold the metadata specifications of a class.

    It is an MSONable class well to allow for easy serialization.

    This class should be used to record the specifications used to generate data, for example
    in the sampling metadata dictionary of the SampleContainer.

    Monte Carlo helper classes include:
        * MCKernels
        * MCUshers
        * MCBias
    """

    def __init__(self, cls_name=None, **kwargs):
        """Initialize the namespace.

        # TODO without default cls_name=None deepcopy fails.

        Args:
            cls_name (str):
                The name of the class for which specifications are being
                recorded.
            **kwargs:
                keyword arguments specifications.
        """
        kwargs["cls_name"] = cls_name
        super().__init__(**kwargs)

    def as_dict(self):
        """Return copy of underlying dictionary."""
        d = self.__dict__.copy()
        for k, v in d.items():
            if isinstance(v, MSONable):
                d[k] = v.as_dict()
            else:
                d[k] = jsanitize(v)
        d.update(
            {"@class": self.__class__.__name__, "@module": self.__class__.__module__}
        )
        return d

    @classmethod
    def from_dict(cls, d):
        """Initialize from dictionary."""
        # try to recreate any MSONables in the dictionary
        decoded = {
            k: MontyDecoder().process_decoded(v)
            for k, v in d.items()
            if not k.startswith("@")
        }
        return cls(**decoded)



"""Definitions of c structs

The structs are used to hold data needed to compute correlation and cluster
interaction vectors.
"""

__author__ = "Luis Barroso-Luque"

cimport numpy as np


cdef struct IntArray1D:
    np.int32_t* data
    int size


cdef struct IntArray2D:
    np.int32_t* data
    int size_r
    int size_c


cdef struct FloatArray1D:
    np.float64_t* data
    int size


cdef struct FloatArray2D:
    np.float64_t* data
    int size_r
    int size_c


cdef struct OrbitC:
    int id
    int bit_id
    IntArray1D tensor_indices
    FloatArray2D correlation_tensors



"""Cython based utilities for computing correlations and cluster interactions."""


def get_orbit_data(orbits):
    """Gather orbit data necessary to create a ClusterSpaceEvaluator."""
    orbit_data = tuple(
        (
            orbit.id,
            orbit.bit_id,
            orbit.flat_correlation_tensors,
            orbit.flat_tensor_indices,
        )
        for orbit in orbits
    )
    return orbit_data



"""Implementation of MCMC transition kernel classes.

A kernel essentially is an implementation of the MCMC algorithm that is used
to generate states for sampling an MCMC chain.
"""

from smol.moca.kernel.base import MCKernelInterface
from smol.moca.kernel.metropolis import Metropolis, MulticellMetropolis
from smol.moca.kernel.random import UniformlyRandom
from smol.moca.kernel.wanglandau import WangLandau
from smol.utils.class_utils import (
    class_name_from_str,
    derived_class_factory,
    get_subclasses_str,
)

__all__ = [
    "Metropolis",
    "MulticellMetropolis",
    "WangLandau",
    "UniformlyRandom",
    "mckernel_factory",
]


def available_mckernels() -> tuple[str]:
    """Return names of available MCMC kernels.

    Returns:
        tuple[str]: names of available MCMC kernels
    """
    return get_subclasses_str(MCKernelInterface)


def mckernel_factory(kernel_type, ensemble, step_type, *args, **kwargs):
    """Get a MCMC Kernel from string name.

    Args:
        kernel_type (str):
            string specifying kernel type to instantiate.
        ensemble (Ensemble)
            an Ensemble object to create the MCMC kernel from.
        step_type (str):
            string specifying the proposal type (i.e. key for MCUsher type)
        *args:
            positional arguments passed to class constructor
        **kwargs:
            keyword arguments passed to class constructor

    Returns:
        MCKernel: instance of derived class.
    """
    kernel_name = class_name_from_str(kernel_type)
    return derived_class_factory(
        kernel_name, MCKernelInterface, ensemble, step_type, *args, **kwargs
    )



import os

import pytest

from smol.utils import _openmp_helpers
from smol.utils.cluster import numthreads


class Dummy:
    num_threads = numthreads.SetNumThreads("_multithreaded_dummy")

    def __init__(self, num_threads=None):
        self._multithreaded_dummy = MultithreadedDummy()
        self.num_threads = num_threads


class MultithreadedDummy:
    def __init__(self):
        self.num_threads = 1


def test_openmp_numthreads(monkeypatch):
    monkeypatch.delenv("OMP_NUM_THREADS", raising=False)
    dummy = Dummy()

    if _openmp_helpers._openmp_enabled():
        assert _openmp_helpers._openmp_effective_numthreads(2) == 2
        assert _openmp_helpers._openmp_effective_numthreads() == min(
            os.cpu_count(), _openmp_helpers._openmp_get_max_threads()
        )
        assert _openmp_helpers._openmp_effective_numthreads(-1) == min(
            os.cpu_count(), _openmp_helpers._openmp_get_max_threads()
        )
        assert (
            _openmp_helpers._openmp_effective_numthreads(-2)
            == min(os.cpu_count(), _openmp_helpers._openmp_get_max_threads()) - 1
        )
        assert dummy.num_threads == 2
        assert Dummy(num_threads=1).num_threads == 1
        with pytest.warns(UserWarning):
            dummy.num_threads = (
                min(os.cpu_count(), _openmp_helpers._openmp_get_max_threads()) + 1
            )
        assert dummy.num_threads == min(
            os.cpu_count(), _openmp_helpers._openmp_get_max_threads()
        )
    else:
        assert _openmp_helpers._openmp_effective_numthreads(2) == 1
        assert _openmp_helpers._openmp_effective_numthreads() == 1
        assert _openmp_helpers._openmp_effective_numthreads(-1) == 1
        assert dummy.num_threads == 1



import os
from itertools import product

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core.composition import Composition

from smol.cofe.space.domain import SiteSpace
from smol.moca import SampleContainer
from smol.moca.sublattice import Sublattice
from smol.moca.trace import Trace
from tests.utils import assert_msonable

NSAMPLES = 100
# compositions will depend on number of sites and num species in the
# sublattices created in the fixture, if changing make sure it works out.
SUBLATTICE_COMPOSITIONS = [1.0 / 3.0, 1.0 / 2.0]


@pytest.fixture(params=[1, 5])
def container(request, single_sgc_ensemble, rng):
    trace = Trace(
        occupancy=np.zeros(
            (0, request.param, single_sgc_ensemble.num_sites), dtype=int
        ),
        features=rng.random(
            (0, request.param, len(single_sgc_ensemble.natural_parameters))
        ),
        enthalpy=np.ones((0, request.param, 1)),
        temperature=np.zeros((0, request.param, 1)),
        accepted=np.zeros((0, request.param, 1), dtype=bool),
    )
    sample_container = SampleContainer(
        single_sgc_ensemble,
        sample_trace=trace,
        sampling_metadata=single_sgc_ensemble.thermo_boundaries,
    )
    yield sample_container
    sample_container.clear()


@pytest.fixture
def fake_traces(container, rng):
    nwalkers = container.shape[0]
    occus = np.empty((NSAMPLES, nwalkers, container.shape[-1]))
    enths = -5 * np.ones((NSAMPLES, nwalkers, 1))
    temps = -300.56 * np.ones((NSAMPLES, nwalkers, 1))
    feats = np.zeros((NSAMPLES, nwalkers, len(container.natural_parameters)))
    accepted = rng.integers(2, size=(NSAMPLES, nwalkers, 1))
    sites1 = container.sublattices[0].sites
    sites2 = container.sublattices[1].sites
    for i in range(NSAMPLES):
        for j in range(nwalkers):
            # make occupancy compositions (1/3, 1/3, 1/3) & (1/2, 1/2)
            size = int(SUBLATTICE_COMPOSITIONS[0] * len(sites1))
            s = rng.choice(sites1, size=size, replace=False)
            occus[i, j, s] = 0
            s1 = rng.choice(np.setdiff1d(sites1, s), size=size, replace=False)
            occus[i, j, s1] = 1
            s2 = np.setdiff1d(sites1, np.append(s, s1))
            occus[i, j, s2] = 2
            size = int(SUBLATTICE_COMPOSITIONS[1] * len(sites2))
            s = rng.choice(sites2, size=size, replace=False)
            occus[i, j, s] = 0
            occus[i, j, np.setdiff1d(sites2, s)] = 1
            # first and last feature real fake
            feats[i, j, [0, -1]] = 2.5
    traces = [
        Trace(
            occupancy=occus[i],
            features=feats[i],
            enthalpy=enths[i],
            accepted=accepted[i],
            temperature=temps[i],
        )
        for i in range(len(accepted))
    ]
    return traces


def add_samples(sample_container, fake_traces, thinned_by=1):
    sample_container.allocate(len(fake_traces))
    for trace in fake_traces:
        sample_container.save_sampled_trace(trace, thinned_by=thinned_by)


def test_allocate_and_save(container, fake_traces, rng):
    nwalkers = container.shape[0]
    nsites = container.shape[-1]
    assert len(container) == 0
    assert container._trace.occupancy.shape == (0, nwalkers, nsites)
    assert container._trace.enthalpy.shape == (0, nwalkers, 1)
    assert container._trace.accepted.shape == (0, nwalkers, 1)

    container.allocate(NSAMPLES)
    assert len(container) == 0
    assert container._trace.occupancy.shape == (NSAMPLES, nwalkers, nsites)
    assert container._trace.enthalpy.shape == (NSAMPLES, nwalkers, 1)
    assert container._trace.accepted.shape == (NSAMPLES, nwalkers, 1)
    container.clear()

    add_samples(container, fake_traces)
    assert len(container) == NSAMPLES
    assert container._total_steps == NSAMPLES
    container.clear()

    thinned = rng.integers(50)
    add_samples(container, fake_traces, thinned_by=thinned)
    assert len(container) == NSAMPLES
    assert container._total_steps == thinned * NSAMPLES
    container.clear()


@pytest.mark.parametrize("discard, thin", product((0, 10), (1, 5)))
def test_get_sampled_values(container, fake_traces, discard, thin):
    nsites = container.shape[-1]
    add_samples(container, fake_traces)
    nat_params = container.natural_parameters
    sublattices = container.sublattices
    nw = container.shape[0]
    # get default flatted values
    accepted = container._trace.accepted
    nsamples = (NSAMPLES - discard) // thin
    expected = (
        accepted[discard:].sum(axis=0) / (container._total_steps - discard)
    ).mean()
    assert container.sampling_efficiency(discard=discard) == expected
    assert container.get_occupancies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
        nsites,
    )
    assert container.get_feature_vectors(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
        len(nat_params),
    )
    assert container.get_enthalpies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
    )
    assert container.get_energies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        c = container.get_sublattice_compositions(
            sublattice, discard=discard, thin_by=thin
        )
        if len(sublattice.species) == 1:  # single species sublattice
            assert c.shape == (nsamples * nw,)
        else:
            assert c.shape == (nsamples * nw, len(sublattice.species))
        npt.assert_array_equal(c, comp * np.ones_like(c))

    # get non flattened values
    npt.assert_array_equal(
        container.sampling_efficiency(discard=discard, flat=False),
        (accepted[discard:,].sum(axis=0) / (container._total_steps - discard)),
    )
    assert container.get_occupancies(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, nsites)
    assert container.get_feature_vectors(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, len(nat_params))
    assert container.get_enthalpies(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, 1)
    assert container.get_energies(discard=discard, thin_by=thin, flat=False).shape == (
        nsamples,
        nw,
        1,
    )

    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        c = container.get_sublattice_compositions(
            sublattice, discard=discard, thin_by=thin, flat=False
        )
        assert c.shape == (nsamples, nw, len(sublattice.species))
        npt.assert_array_equal(c, comp * np.ones_like(c))

    with pytest.raises(ValueError):
        sublattice = Sublattice(SiteSpace(Composition({"foo": 7})), np.zeros(10))
        container.get_sublattice_compositions(sublattice)


@pytest.mark.parametrize("discard, thin", product((0, 10), (1, 5)))
def test_means_and_variances(container, fake_traces, discard, thin):
    add_samples(container, fake_traces)
    sublattices = container.sublattices
    nw = container.shape[0]
    assert container.mean_enthalpy(discard=discard, thin_by=thin) == -5.0
    assert container.enthalpy_variance(discard=discard, thin_by=thin) == 0.0
    assert container.mean_energy(discard=discard, thin_by=thin) == -2.5
    assert container.energy_variance(discard=discard, thin_by=thin) == 0.0
    npt.assert_array_equal(
        container.mean_feature_vector(discard=discard, thin_by=thin),
        [2.5]
        + (len(container.natural_parameters) - 2)
        * [
            0,
        ]
        + [2.5],
    )
    npt.assert_array_equal(
        container.feature_vector_variance(discard=discard, thin_by=thin),
        len(container.natural_parameters)
        * [
            0,
        ],
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        npt.assert_array_almost_equal(
            container.mean_sublattice_composition(
                sublattice, discard=discard, thin_by=thin
            ),
            len(sublattice.species)
            * [
                comp,
            ],
        )
        npt.assert_array_almost_equal(
            container.sublattice_composition_variance(
                sublattice, discard=discard, thin_by=thin
            ),
            len(sublattice.species)
            * [
                0,
            ],
        )

    # without flattening
    npt.assert_array_equal(
        container.mean_enthalpy(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [-5.0]]).T,
    )
    npt.assert_array_equal(
        container.enthalpy_variance(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [0.0]]).T,
    )
    npt.assert_array_equal(
        container.mean_energy(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [-2.5]]).T,
    )
    npt.assert_array_equal(
        container.energy_variance(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [0.0]]).T,
    )
    npt.assert_array_equal(
        container.mean_feature_vector(discard=discard, thin_by=thin, flat=False),
        nw
        * [
            [2.5]
            + (len(container.natural_parameters) - 2)
            * [
                0.0,
            ]
            + [2.5]
        ],
    )
    npt.assert_array_equal(
        container.feature_vector_variance(discard=discard, thin_by=thin, flat=False),
        nw
        * [
            len(container.natural_parameters)
            * [
                0.0,
            ]
        ],
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        npt.assert_array_almost_equal(
            container.mean_sublattice_composition(
                sublattice, discard=discard, thin_by=thin, flat=False
            ),
            nw * [len(sublattice.species) * [comp]],
        )
        npt.assert_array_almost_equal(
            container.sublattice_composition_variance(
                sublattice, discard=discard, thin_by=thin, flat=False
            ),
            nw * [len(sublattice.species) * [0]],
        )


def test_get_mins(container, fake_traces, rng):
    add_samples(container, fake_traces)
    i = rng.choice(range(NSAMPLES))
    nwalkers = container.shape[0]
    container._trace.enthalpy[i, :] = -11
    container._trace.features[i, :, 0] = 5.0
    assert container.get_minimum_enthalpy() == -11.0
    assert container.get_minimum_energy() == -5.0
    npt.assert_array_equal(
        container.get_minimum_enthalpy(flat=False), np.array([nwalkers * [-11.0]]).T
    )
    npt.assert_array_equal(
        container.get_minimum_energy(flat=False), np.array([nwalkers * [-5.0]]).T
    )
    npt.assert_array_equal(
        container.get_minimum_enthalpy_occupancy(flat=False),
        container._trace.occupancy[i],
    )
    npt.assert_array_equal(
        container.get_minimum_energy_occupancy(flat=False),
        container._trace.occupancy[i],
    )


def test_msonable(container, fake_traces):
    # fails for empty container with nwalkers > 1
    # since the _chain.tolist turns the empty array to an empty list and so
    # the shape is lost, but dont think anyone really cares about an empty
    # container....
    # d = container.as_dict()
    # cntr = container.from_dict(d)
    # assert cntr.shape == container.shape

    add_samples(container, fake_traces)
    d = container.as_dict()
    cntr = container.from_dict(d)
    assert cntr.shape == container.shape
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    assert_msonable(cntr)


def test_hdf5(container, fake_traces, tmpdir):
    add_samples(container, fake_traces)
    file_path = os.path.join(tmpdir, "test.h5")
    container.to_hdf5(file_path)
    cntr = SampleContainer.from_hdf5(file_path)
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    npt.assert_array_equal(container.get_energies(), cntr.get_energies())
    os.remove(file_path)


@pytest.mark.parametrize("mode", [False, True])
def test_flush_to_hdf5(container, fake_traces, mode, tmpdir):
    # deepcopy does not work with Cython extensions with nontrivial __cinit__
    # and creating a new container from dict of an empty one loses shape information
    # so we add samples first and then create the flushed container, and clear it
    add_samples(container, fake_traces)
    flushed_container = SampleContainer.from_dict(container.as_dict())
    flushed_container.clear()

    file_path = os.path.join(tmpdir, "test.h5")
    chunk = len(fake_traces) // 4
    flushed_container.allocate(chunk)
    backend = flushed_container.get_backend(file_path, len(fake_traces), swmr_mode=mode)

    start = 0
    for _ in range(4):
        for i in range(start, start + chunk):
            flushed_container.save_sampled_trace(fake_traces[i], thinned_by=1)
        assert flushed_container._trace.occupancy.shape[0] == chunk
        flushed_container.flush_to_backend(backend)
        start += chunk

    if mode is False:
        backend.close()

    assert flushed_container._trace.occupancy.shape[0] == chunk
    assert flushed_container.num_samples == 0
    cntr = SampleContainer.from_hdf5(file_path)
    assert cntr.num_samples == container.num_samples
    assert cntr.total_mc_steps == container.total_mc_steps
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    npt.assert_array_equal(container.get_energies(), cntr.get_energies())

    if mode is True:
        backend.close()

    # now get it again, and make sure more space is allocated
    add_samples(container, fake_traces)
    backend = flushed_container.get_backend(file_path, len(fake_traces))
    assert backend["trace"].attrs["nsamples"] == len(fake_traces)
    assert len(backend["trace"]["occupancy"]) == 2 * len(fake_traces)

    start = 0
    for _ in range(4):
        for i in range(start, start + chunk):
            flushed_container.save_sampled_trace(fake_traces[i], thinned_by=1)
        assert flushed_container._trace.occupancy.shape[0] == chunk
        flushed_container.flush_to_backend(backend)
        start += chunk

    assert flushed_container._trace.occupancy.shape[0] == chunk
    assert flushed_container.num_samples == 0
    cntr = SampleContainer.from_hdf5(file_path)
    assert cntr.num_samples == container.num_samples
    assert cntr.total_mc_steps == container.total_mc_steps
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    npt.assert_array_equal(container.get_energies(), cntr.get_energies())

    backend.close()

    os.remove(file_path)


def test_get_sampled_structures(container, fake_traces, rng):
    # not the best test here, but it will do for now...
    add_samples(container, fake_traces)
    ids = rng.choice(range(container.num_samples), size=5)

    # check for flattened chains
    structures = container.get_sampled_structures(ids, flat=True)
    for i, struct in zip(ids, structures):
        occu = container.get_occupancies()[i]
        assert container.ensemble.processor.structure_from_occupancy(occu) == struct

    # check for unflattened chains
    structures_list = container.get_sampled_structures(ids, flat=False)
    for i, structures in zip(ids, structures_list):
        occus = container.get_occupancies(flat=False)[i]
        for occu, struct in zip(occus, structures):
            assert container.ensemble.processor.structure_from_occupancy(occu) == struct


def test_allocate_vacuum(container, fake_traces):
    add_samples(container, fake_traces)
    n_samples = len(container)

    container.allocate(100)
    assert len(container) == n_samples
    assert container._trace.occupancy.shape[0] == n_samples + 100

    container.vacuum()
    assert len(container) == n_samples
    assert container._trace.occupancy.shape[0] == n_samples



"""A few general utilities functions."""

__author__ = "Luis Barroso-Luque"

import inspect
import re
from typing import Any, Dict


def class_name_from_str(class_str):
    """Return a class name based on given string.

    Assumes all class names are properly named with Camel Caps.

    Used to generalize how classes can be identified by also allowing
    capitalized words in class name to be separated by a hyphen and also
    allowing capitalization to be ignored.

    i.e. all of the following are valid strings for class TheBestClass:
    'TheBestClass', 'The-Best-Class', 'the-best-class', 'The-best-Class', etc

    Args:
        class_str (str):
            string identifying a class name.
    Returns: str
        class name in camel caps
    """
    return "".join(
        [
            s.capitalize()
            for sub in class_str.split("-")
            for s in re.findall("[a-zA-Z][^A-Z]*", sub)
        ]
    )


def derived_class_factory(
    class_name: str, base_class: object, *args: Any, **kwargs: Any
) -> object:
    """Return an instance of derived class from a given basis class.

    Args:
        class_name (str):
            name of class
        base_class (object):
            base class of derived class sought
        *args:
            positional arguments for class constructor
        **kwargs:
            keyword arguments for class constructor

    Returns:
        object: instance of class with corresponding constructor args, kwargs
    """
    try:
        derived_class = get_subclasses(base_class)[class_name]
        instance = derived_class(*args, **kwargs)
    except KeyError as key_error:
        raise NotImplementedError(f"{class_name} is not implemented.") from key_error
    return instance


def get_subclasses(base_class: object) -> Dict[str, object]:
    """Get all non-abstract subclasses of a class.

    Gets all non-abstract classes that inherit from the given base class in
    a module.
    """
    sub_classes = {}
    for sub_class in base_class.__subclasses__():
        sub_classes.update(get_subclasses(sub_class))

        if not inspect.isabstract(sub_class):
            sub_classes[sub_class.__name__] = sub_class

    return sub_classes


def get_subclasses_str(
    base_class: object, lower: bool = True, split: bool = True
) -> tuple[str]:
    """Get names of all non-abstract subclasses of a class.

    Gets all names of non-abstract classes that inherit from the given base class in
    a module.

    Args:
        base_class (object): base class to get subclasses of
        lower (bool): whether to return names in lower case
        split (bool): whether to split the class name and add hyphens between words
    """
    names = get_subclasses(base_class).keys()

    if split:
        names = tuple("-".join(re.findall("[A-Z][^A-Z]*", name)) for name in names)
    if lower:
        names = tuple(name.lower() for name in names)
    return names



"""Evaluator cython extension type to for fast computation of correlation vectors."""

__author__ = "Luis Barroso-Luque"

cimport numpy as np

import cython

from smol.utils.cluster.container cimport (
    FloatArray1DContainer,
    IntArray2DContainer,
    OrbitContainer,
)


@cython.final
cdef class ClusterSpaceEvaluator(OrbitContainer):
    cdef int num_corr  # total number of correlation functions
    cdef int num_orbits  # total number of orbits
    cdef public double offset  # offset for the interaction tensor
    cdef FloatArray1DContainer cluster_interactions
    cdef public int num_threads

    cpdef public void reset_data(
            self,
            tuple orbit_data,
            int num_orbits,
            int num_corr_functions,
    )

    cpdef public void set_cluster_interactions(
            self, tuple cluster_interaction_tensors, double offset
    )

    cpdef np.ndarray[np.float64_t, ndim=1] correlations_from_occupancy(
            self,
            const np.int32_t[::1] occu,
            IntArray2DContainer cluster_indices,
    )

    cpdef np.ndarray[np.float64_t, ndim=1] interactions_from_occupancy(
            self,
            const np.int32_t[::1] occu,
            IntArray2DContainer cluster_indices,
    )

    cpdef np.ndarray[np.float64_t, ndim=1] delta_correlations_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] cluster_ratio,
            IntArray2DContainer cluster_indices)

    cpdef np.ndarray[np.float64_t, ndim=1] delta_interactions_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] cluster_ratio,
            IntArray2DContainer cluster_indices
    )

    cpdef np.ndarray[np.float64_t, ndim=1] corr_distances_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] ref_corr_vector,
            IntArray2DContainer cluster_indices
    )

    cpdef np.ndarray[np.float64_t, ndim=1] interaction_distances_from_occupancies(
            self,
            const np.int32_t[::1] occu_f,
            const np.int32_t[::1] occu_i,
            const np.float64_t[::1] ref_interaction_vector,
            IntArray2DContainer cluster_indices
    )



"""Implementation of base sampler classes.

A sampler essentially is an implementation of the MCMC algorithm that is used
by the corresponding ensemble to generate Monte Carlo samples.
"""

__author__ = "Luis Barroso-Luque"

import os
from datetime import datetime
from warnings import warn

import numpy as np

from smol.moca.kernel import mckernel_factory
from smol.moca.kernel.base import MulticellKernel
from smol.moca.sampler.container import SampleContainer
from smol.moca.trace import Trace
from smol.utils.progressbar import progress_bar


class Sampler:
    """
    A sampler is used to run MCMC sampling simulations.

    The specific MCMC algorithm is defined by the given MCKernel.
    The default will use a simple Metropolis random walk kernel.
    """

    def __init__(self, kernels, container):
        """Initialize BaseSampler.

        It is recommended to initialize a sampler with the from_ensemble
        method, unless you need more control over the sampling kernel,
        mcusher, etc.

        Args:
            kernels ([MCKernel]):
                a list of MCKernel instances for each walker.
                .. note::
                    we only support the same type of kernel and the same ensemble for
                    all walkers.
            container (SampleContainer):
                a sampler container to store samples.
        """
        self._kernels = kernels
        self._container = container

        #  Save kernel specifications  #
        self._container.metadata.kernels = [kernel.spec for kernel in kernels]

    @classmethod
    def from_ensemble(
        cls,
        ensemble,
        *args,
        step_type=None,
        kernel_type=None,
        seeds=None,
        nwalkers=1,
        **kwargs,
    ):
        """
        Create a sampler based on an Ensemble instance.

        This is the easier way to set up a Sampler. This will
        automatically populate and create an appropriate SampleContainer.

        Args:
            ensemble (Ensemble):
                an Ensemble class to obtain sample probabilities from.
            step_type (str): optional
                type of step to run MCMC with. If not given the default depends on
                whether chemical potentials are defined.
            *args:
                positional arguments to pass to the MCKernel constructor.
                More often than not you want to specify the temperature!
            kernel_type (str): optional
                string specifying the specific MCMC transition kernel. This
                represents the underlying MC algorithm. Currently only
                Metropolis is supported.
            seeds ([int]): optional
                seed for the PRNG in each kernel.
            nwalkers (int): optional
                number of walkers/chains to sampler. Default is 1. More than 1
                is still experimental...
            **kwargs:
                keyword arguments to pass to the MCKernel constructor.
                More often than not you want to specify the temperature!

        Returns:
            Sampler
        """
        if step_type is None:
            if (
                hasattr(ensemble, "chemical_potentials")
                and ensemble.chemical_potentials is not None
            ):
                step_type = "flip"
            else:
                step_type = "swap"

        if kernel_type is None:
            kernel_type = "Metropolis"
        if seeds is not None:
            if len(seeds) != nwalkers:
                raise ValueError("Number of seeds does not match number of kernels!")
        else:
            seeds = [None for _ in range(nwalkers)]

        mckernels = [
            mckernel_factory(
                kernel_type, ensemble, step_type, seed=seed, *args, **kwargs
            )
            for seed in seeds
        ]

        # get a trial trace to initialize sample container trace
        # Enforce int32.
        _trace = mckernels[0].compute_initial_trace(
            np.zeros(ensemble.num_sites, dtype=np.int32)
        )
        sample_trace = Trace(
            **{
                name: np.empty((0, nwalkers, *value.shape), dtype=value.dtype)
                for name, value in _trace.items()
            }
        )

        sampling_metadata = ensemble.thermo_boundaries

        # Container will be initialized to read all sub-lattices,
        # active or not.
        container = SampleContainer(
            ensemble,
            sample_trace,
            sampling_metadata,
        )
        return cls(mckernels, container)

    @property
    def mckernels(self):
        """Get the underlying ensemble."""
        return self._kernels

    @property
    def seeds(self):
        """Seed for the random number generator."""
        return [kernel.seed for kernel in self._kernels]

    @property
    def samples(self):
        """Get the SampleContainer."""
        return self._container

    def efficiency(self, discard=0, flat=True):
        """Get the efficiency of MCMC sampling (accepted/total)."""
        return self.samples.sampling_efficiency(discard=discard, flat=flat)

    def clear_samples(self):
        """Clear samples from sampler container."""
        self.samples.clear()

    def sample(self, nsteps, initial_occupancies, thin_by=1, progress=False):
        """Generate MCMC samples.

        Yield a sampler state every `thin_by` iterations. A state is give by
        a tuple of (occupancies, features, enthalpy)

        Args:
            nsteps (int):
                number of iterations to run.
            initial_occupancies (ndarray):
                array of occupancies
            thin_by (int): optional
                number to thin iterations by and provide samples.
            progress (bool):
                If true will show a progress bar.

        Yields:
            tuple: accepted, occupancies, features change, enthalpies change
        """
        if nsteps % thin_by != 0:
            warn(
                f"The number of steps {nsteps} is not a multiple of thin_by "
                f" {thin_by}. The last {nsteps % thin_by} will be ignored.",
                category=RuntimeWarning,
            )

        occupancies, trace = self.setup_sample(initial_occupancies)
        # Initialise progress bar
        chains, nsites = self.samples.shape
        desc = f"Sampling {chains} chain(s) from a cell with {nsites} sites"
        with progress_bar(progress, total=nsteps, description=desc) as p_bar:
            for _ in range(nsteps // thin_by):
                for _ in range(thin_by):
                    # the occupancies are modified in place at each step
                    for i, strace in enumerate(self._single_step(occupancies)):
                        for name, value in strace.items():
                            val = getattr(trace, name)
                            val[i] = value
                            # this will mess up recording values for > 1 walkers
                            # setattr(trace, name, value)
                        if strace.accepted:
                            for name, delta_val in strace.delta_trace.items():
                                val = getattr(trace, name)
                                val[i] += delta_val
                    p_bar.update()
                # yield copies
                yield trace

    def run(
        self,
        nsteps,
        initial_occupancies=None,
        thin_by=1,
        progress=False,
        stream_chunk=0,
        stream_file=None,
        keep_last_chunk=False,
        swmr_mode=False,
    ):
        """Run an MCMC sampling simulation.

        This will run and save the samples every thin_by into a
        SampleContainer.

        Args:
            nsteps (int):
                number of total MC steps.
            initial_occupancies (ndarray):
                array of occupancies. If None, the last sampler will be taken.
                You should only provide this the first time you call run. If
                you want to reset then you should call reset before to start
                a fresh run.
            thin_by (int): optional
                the amount to thin by for saving samples.
            progress (bool): optional
                if true will show a progress bar.
            stream_chunk (int): optional
                chunk of samples to stream into a file. If > 0 samples will
                be flushed to backend file in stream_chucks
            stream_file (str): optional
                file name to use as backend. If file already exists will try
                to append to datasets. If not given will create a new file.
            keep_last_chunk (bool): optional
                if True will keep the last chunk of samples in memory, but will reset
                the sampler otherwise. This is useful if the last occupancies are
                needed to start another run (ie for simulated annealing with streaming)
            swmr_mode (bool): optional
                if true allows to read file from other processes. Single Writer
                Multiple Readers.
        """
        if initial_occupancies is None:
            try:
                initial_occupancies = self.samples.get_occupancies(flat=False)[-1]
                # auxiliary states from kernels should be set here
            except IndexError as index_error:
                raise RuntimeError(
                    "There are no saved samples to obtain the initial occupancies."
                    "These must be provided."
                ) from index_error
        elif self.samples.num_samples > 0:
            warn(
                "Initial occupancies where provided with a pre-existing set of samples."
                "\n Make real sure that is what you want. "
                "If not, reset the samples in the sampler.",
                RuntimeWarning,
            )

        if stream_chunk > 0:
            if stream_file is None:
                now = datetime.now()
                file_name = "moca-samples-" + now.strftime("%Y-%m-%d-%H%M%S%f")
                stream_file = os.path.join(os.getcwd(), file_name + ".h5")
            backend = self.samples.get_backend(
                stream_file, nsteps // thin_by, swmr_mode=swmr_mode
            )
            # allocate memory only if there is None available
            if len(self.samples.get_occupancies()) == 0:
                self.samples.allocate(stream_chunk)
        else:
            backend = None
            self.samples.allocate(nsteps // thin_by)

        for i, trace in enumerate(
            self.sample(nsteps, initial_occupancies, thin_by=thin_by, progress=progress)
        ):
            self.samples.save_sampled_trace(trace, thinned_by=thin_by)
            if backend is not None and (i + 1) % stream_chunk == 0:
                self.samples.flush_to_backend(backend)

        if backend is not None:
            backend.close()
            # Only clear samples if requested.
            if keep_last_chunk is False:
                self.clear_samples()

        # A checkpoint of aux states should be saved to container here.
        # Note that to save any general "state" we will need to make sure it is
        # properly serializable to save as json and also to save in h5py

    def anneal(
        self,
        temperatures,
        mcmc_steps,
        initial_occupancies=None,
        thin_by=1,
        progress=False,
        stream_chunk=0,
        stream_file=None,
        swmr_mode=True,
    ):
        """Carry out a simulated annealing procedure.

        Will MC sampling for each temperature for the specified number of steps,
        taking the last sampled configuration at each temperature as the starting
        configuration for the next temperature.

        Everything is saved to the same SampleContainer, or if streaming, saved to the
        same file. To save each temperature in a different container, simply run
        a similar "for loop" creating a new sampler at each temperature.

        .. note::
            This method is only available for Samplers with a ThermalKernel.

        Args:
            temperatures (Sequence):
               sequence of temperatures to anneal, should be strictly
               decreasing.
            mcmc_steps (int):
               number of Monte Carlo steps to run at each temperature.
            initial_occupancies (ndarray):
                array of occupancies. If None, the last sample will be taken.
                You should only provide this the first time you call run.
            thin_by (int): optional
                amount to thin by for saving samples.
            progress (bool):
                if true will show a progress bar.
            stream_chunk (int): optional
                chunk of samples to stream into a file. If > 0 samples will
                be flushed to backend file in stream_chucks
            stream_file (str): optional
                file name to use as backend. If file already exists will try
                to append to datasets. If not given will create a new file.
            swmr_mode (bool): optional
                if true allows to read file from other processes. Single Writer
                Multiple Readers.
        """
        if temperatures[0] < temperatures[-1]:
            raise ValueError(
                "End temperature is greater than start "
                f"temperature {temperatures[-1]:.2f} > {temperatures[0]:.2f}."
            )

        # initialize for first temperature.
        for kernel in self._kernels:
            kernel.temperature = temperatures[0]
        self.run(
            mcmc_steps,
            initial_occupancies=initial_occupancies,
            thin_by=thin_by,
            progress=progress,
            stream_chunk=stream_chunk,
            stream_file=stream_file,
            swmr_mode=swmr_mode,
            keep_last_chunk=True,
        )
        for temperature in temperatures[1:]:
            for kernel in self._kernels:
                kernel.temperature = temperature
            self.run(
                mcmc_steps,
                thin_by=thin_by,
                progress=progress,
                stream_chunk=stream_chunk,
                stream_file=stream_file,
                swmr_mode=swmr_mode,
                keep_last_chunk=True,
            )

        # If streaming to file was done then clear samplers now.
        if stream_chunk > 0:
            self.clear_samples()

    def setup_sample(self, initial_occupancies):
        """Copy and reshape occupancies and compute initial trace for sampling.

        Args:
            initial_occupancies (ndarray):
                initial occupancies for sampling.

        Returns: (occupancies, trace)
            occupancies (ndarray):
                initial occupancies for sampling.
            trace (Trace):
                initial trace for sampling.
        """
        # TODO check that initial states are independent if num_walkers > 1
        # TODO make samplers with single chain, multiple and multiprocess
        occupancies = initial_occupancies.copy()
        if occupancies.shape != self.samples.shape:
            occupancies = self._reshape_occu(occupancies)
        # Enforce int32 here at the beginning of simulation.
        # This will ensure that all following steps are in int32 format.
        occupancies = occupancies.astype(np.int32)

        # set up any auxiliary states from initial occupancies
        selected_occus = []
        for kernel, occupancy in zip(self._kernels, occupancies):
            kernel.set_aux_state(occupancy)
            # select the current kernel occu
            if isinstance(kernel, MulticellKernel):
                if occupancies.ndim == 3 and occupancies.shape[1] == len(
                    kernel.mckernels
                ):
                    selected_occus.append(occupancy[kernel.trace.kernel_index])
        if len(selected_occus) > 0:
            occupancies = np.vstack(selected_occus)

        # get initial traces and stack them
        trace = Trace()
        traces = list(
            map(
                lambda kernel, occu: kernel.compute_initial_trace(occu),
                self._kernels,
                occupancies,
            )
        )
        for name in traces[0].names:
            stack = np.stack([getattr(tr, name) for tr in traces], axis=0)
            setattr(trace, name, stack)

        return occupancies, trace

    def _single_step(self, occupancies):
        """Do a single step for all kernels."""
        # Don't enforce int32 here as it may increase overhead.
        for kernel, occupancy in zip(self._kernels, occupancies):
            yield kernel.single_step(occupancy)

    def _reshape_occu(self, occupancies):
        """Reshape occupancies for the single walker case."""
        # check if this is only a single walker.
        if len(occupancies.shape) == 1 and self.samples.shape[0] == 1:
            occupancies = np.reshape(occupancies, (1, len(occupancies)))
        elif any(isinstance(kernel, MulticellKernel) for kernel in self._kernels):
            # reshape for multicell kernels
            occupancies = np.reshape(occupancies, (1, *occupancies.shape))
        else:
            raise AttributeError(
                "The given initial occcupancies have incompompatible dimensions. Shape should"
                f" be {self.samples.shape}."
            )
        return occupancies


# TODO potential define two _sample functions serial/parallel that get called
#  in one sampler function.



#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False
#cython: initializedcheck=False
#cython: cdivision=True

"""
General cython utility functions to calculate correlation vectors and local
changes in correlation vectors a tad bit faster than pure python.
"""

__author__ = "Luis Barroso-Luque, William D. Richards"

import numpy as np
cimport numpy as np


cpdef corr_from_occupancy(const np.int32_t[::1] occu,
                          const int num_corr_functions,
                          list orbit_list):
    """Computes the correlation vector for a given encoded occupancy string.

    Args:
        occu (ndarray):
            encoded occupancy array
        num_corr_functions (int):
            total number of bit orderings in expansion.
        orbit_list:
            Information of all orbits.
            (orbit id, flat tensor index array, flat correlation tensor,
             site indices of clusters)

    Returns: array
        correlation vector for given occupancy
    """
    cdef int i, j, n, m, I, J, M, index
    cdef double p
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[:, ::1] corr_tensors
    out = np.zeros(num_corr_functions)
    cdef np.float64_t[:] o_view = out
    o_view[0] = 1  # empty cluster

    for n, tensor_indices, corr_tensors, indices in orbit_list:
        M = corr_tensors.shape[0]  # index of bit combos
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        for m in range(M):
            p = 0
            for i in range(I):
                index = 0
                for j in range(J):
                    index += tensor_indices[j] * occu[indices[i, j]]
                p += corr_tensors[m, index]
            o_view[n] = p / I
            n += 1
    return out


cpdef delta_corr_single_flip(const np.int32_t[::1] occu_f,
                             const np.int32_t[::1] occu_i,
                             const int num_corr_functions,
                             list site_orbit_list):
    """Computes the correlation difference between two occupancy arrays.

    Args:
        occu_f (ndarray):
            encoded occupancy array with flip
        occu_i (ndarray):
            encoded occupancy array without flip
        num_corr_functions (int):
            total number of bit orderings in expansion.
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (orbit id, cluster ratio, flat tensor index array,
             flat correlation tensor, site indices of clusters)

    Returns:
        ndarray: correlation vector difference
    """
    cdef int i, j, n, m, I, J, M, ind_i, ind_f
    cdef double p, ratio
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[:, ::1] corr_tensors
    out = np.zeros(num_corr_functions)
    cdef np.float64_t[::1] o_view = out

    for n, ratio, tensor_indices, corr_tensors, indices in site_orbit_list:
        M = corr_tensors.shape[0] # index of bit combos
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        for m in range(M):
            p = 0
            for i in range(I):
                ind_i, ind_f = 0, 0
                for j in range(J):
                    ind_i += tensor_indices[j] * occu_i[indices[i, j]]
                    ind_f += tensor_indices[j] * occu_f[indices[i, j]]
                p += (corr_tensors[m, ind_f] - corr_tensors[m, ind_i])
            o_view[n] = p / ratio / I
            n += 1
    return out


cpdef corr_distance_single_flip(const np.int32_t[::1] occu_f,
                                const np.int32_t[::1] occu_i,
                                const np.float64_t[::1] ref_corr_vector,
                                const int num_corr_functions,
                                list orbit_list):
    """Computes the absolute distance of two correlation vectors separated by a single
    flip and a given correlation vector.

    Unfortunately this scales just as bad as computing the full correlation vector.

    Args:
        occu_f (ndarray):
            encoded occupancy array with flip
        occu_i (ndarray):
            encoded occupancy array without flip
        ref_corr_vector (ndarray):
            reference correlation vector
        num_corr_functions (int):
            total number of bit orderings in expansion.
        orbit_list:
            Information of all orbits.
            (orbit id, flat tensor index array, flat correlation tensor,
             site indices of clusters)

    Returns:
        ndarray: 2D with correlation vector distances from reference for each of occu_i
        and occu_f
    """
    cdef int i, j, n, m, I, J, M, ind_i, ind_f
    cdef double p_i, p_f
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[:, ::1] corr_tensors
    out = np.zeros((2, num_corr_functions))
    cdef np.float64_t[:, ::1] o_view = out
    o_view[:, 0] = 0

    for n, tensor_indices, corr_tensors, indices in orbit_list:
        M = corr_tensors.shape[0]  # index of bit combos
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        for m in range(M):
            p_f, p_i = 0, 0
            for i in range(I):
                ind_f, ind_i = 0, 0
                for j in range(J):
                    ind_f += tensor_indices[j] * occu_f[indices[i, j]]
                    ind_i += tensor_indices[j] * occu_i[indices[i, j]]
                p_f += corr_tensors[m, ind_f]
                p_i += corr_tensors[m, ind_i]
            o_view[1, n] = abs(p_f / I - ref_corr_vector[n])
            o_view[0, n] = abs(p_i / I - ref_corr_vector[n])
            n += 1
    return out


cpdef interactions_from_occupancy(const np.int32_t[::1] occu,
                                  const int num_interactions,
                                  const double offset,
                                  list orbit_list):
    """Computes the cluster interaction vector for a given encoded occupancy string.

    Args:
        occu (ndarray):
            encoded occupancy vector
        num_interactions (int):
            total number of cluster interactions (orbits in cluster subspace).
        offset (double):
            eci value for the constant term.
        orbit_list:
            Information of all orbits.
            (flat tensor index array, flat cluster interaction tensor,
             site indices of clusters)
    Returns: array
        cluster interaction vector for given occupancy
    """
    cdef int n, i, j, I, J, index
    cdef double p
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[::1] interaction_tensor
    out = np.zeros(num_interactions)
    cdef np.float64_t[:] o_view = out
    o_view[0] = offset  # empty cluster

    n = 1
    for tensor_indices, interaction_tensor, indices in orbit_list:
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        p = 0
        for i in range(I):
            index = 0
            for j in range(J):
                index += tensor_indices[j] * occu[indices[i, j]]
            p += interaction_tensor[index]
        o_view[n] = p / I
        n += 1

    return out


cpdef delta_interactions_single_flip(const np.int32_t[::1] occu_f,
                                     const np.int32_t[::1] occu_i,
                                     const int num_interactions,
                                     list site_orbit_list):
    """Computes the cluster interaction vector difference between two occupancy
    strings.
    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        num_interactions (int):
            total number of cluster interactions (orbits in cluster subspace).
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (cluster ratio, flat tensor index array, flat cluster interaction tensor,
             site indices of clusters)
    Returns:
        ndarray: cluster interaction vector difference
    """
    cdef int i, j, n, I, J, ind_i, ind_f
    cdef double p, ratio
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[::1] interaction_tensor
    out = np.zeros(num_interactions)
    cdef np.float64_t[::1] o_view = out

    for n, ratio, tensor_indices, interaction_tensor, indices in site_orbit_list:
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        p = 0
        for i in range(I):
            ind_i, ind_f = 0, 0
            for j in range(J):
                ind_i += tensor_indices[j] * occu_i[indices[i, j]]
                ind_f += tensor_indices[j] * occu_f[indices[i, j]]
            p += (interaction_tensor[ind_f] - interaction_tensor[ind_i])
        o_view[n] = p / ratio / I
    return out


cpdef interaction_distance_single_flip(const np.int32_t[::1] occu_f,
                                       const np.int32_t[::1] occu_i,
                                       const np.float64_t[::1] ref_interaction_vector,
                                       const int num_interactions,
                                       list orbit_list):
    """Computes the absolute distance of two cluster interaction vectors separated by a
    single flip and a given correlation vector.

    Unfortunately this scales just as bad as computing the full interaction vector.

    Args:
        occu_f (ndarray):
            encoded occupancy array with flip
        occu_i (ndarray):
            encoded occupancy array without flip
        ref_interaction_vector (ndarray):
            reference cluster interaction vector
        num_interactions (int):
            total number of cluster interactions (orbits in cluster subspace).
        site_orbit_list:
            Information of all orbits that include the flip site.
            List of tuples each with
            (cluster ratio, flat tensor index array, flat cluster interaction tensor,
             site indices of clusters)

    Returns:
        ndarray: 2D with cluster interaction vector distances from reference for each of
        occu_i and occu_f
    """
    cdef int n, i, j, I, J, ind_i, ind_f
    cdef double p_i, p_f
    cdef const np.int32_t[:, ::1] indices
    cdef const np.int32_t[::1] tensor_indices
    cdef const np.float64_t[::1] interaction_tensor
    out = np.zeros((2, num_interactions))
    cdef np.float64_t[:, ::1] o_view = out
    o_view[:, 0] = 0

    n = 1
    for tensor_indices, interaction_tensor, indices in orbit_list:
        I = indices.shape[0] # cluster index
        J = indices.shape[1] # index within cluster
        p_f, p_i = 0, 0
        for i in range(I):
            ind_f, ind_i = 0, 0
            for j in range(J):
                ind_f += tensor_indices[j] * occu_f[indices[i, j]]
                ind_i += tensor_indices[j] * occu_i[indices[i, j]]
            p_f += interaction_tensor[ind_f]
            p_i += interaction_tensor[ind_i]
        o_view[1, n] = abs(p_f / I - ref_interaction_vector[n])
        o_view[0, n] = abs(p_i / I - ref_interaction_vector[n])
        n += 1
    return out


cpdef delta_ewald_single_flip(const np.int32_t[::1] occu_f,
                              const np.int32_t[::1] occu_i,
                              const np.float64_t[:, ::1] ewald_matrix,
                              const np.int32_t[:, ::1] ewald_indices,
                              const int site_ind):
    """Compute the change in electrostatic interaction energy from a flip.

    Args:
        occu_f (ndarray):
            encoded occupancy array with flip
        occu_i (ndarray):
            encoded occupancy array without flip
        ewald_matrix (ndarray):
            Ewald matrix for electrostatic interactions
        ewald_indices (ndarray):
            2D array of indices corresponding to a specific site occupation
            in the ewald matrix
        site_ind (int):
            site index for site being flipped

    Returns:
        double: electrostatic interaction energy difference
    """
    cdef int i, j, k, add, sub
    cdef bint ok
    cdef double out = 0
    cdef double out_k

    # values of -1 are vacancies and hence don't have ewald indices
    add = ewald_indices[site_ind, occu_f[site_ind]]
    sub = ewald_indices[site_ind, occu_i[site_ind]]

    for k in range(occu_f.shape[0]):
        i = ewald_indices[k, occu_f[k]]
        out_k = 0
        if i != -1 and add != -1:
            if i != add:
                out_k = out_k + 2 * ewald_matrix[i, add]
            else:
                out_k = out_k + ewald_matrix[i, add]

        j = ewald_indices[k, occu_i[k]]
        if j != -1 and sub != -1:
            if j != sub:
                out_k = out_k - 2 * ewald_matrix[j, sub]
            else:
                out_k = out_k - ewald_matrix[j, sub]
        out += out_k
    return out



"""Implementation of SampleContainer class.

A SampleContainer is used to store data from Monte Carlo sampling simulation.
It has some minimimal methods and properties useful to start analyzing the
samples.
"""

__author__ = "Luis Barroso-Luque"

import json
import os
import warnings
from collections import defaultdict

import h5py
import numpy as np
from monty.json import MontyDecoder, MontyEncoder, MSONable, jsanitize

from smol.moca import Ensemble
from smol.moca.metadata import Metadata
from smol.moca.sublattice import Sublattice
from smol.moca.trace import Trace


class SampleContainer(MSONable):
    """A SampleContainer class stores Monte Carlo simulation samples.

    A SampleContainer holds samples and sampling information from an MCMC
    sampling run. It is useful to obtain the raw data and minimal empirical
    properties of the underlying distribution in order to carry out further
    analysis of the MCMC sampling results.

    When getting any value from the provided attributes, the highly repeated
    args are:

        discard (int): optional
            number of samples to discard to obtain the value requested.
        thin_by (int): optional
            use every thin_by sample to obtain the value requested.
        flat (bool): optional
            if more than 1 walker is used, flattening will flatten all
            chains into one. Defaults to True.

    Attributes:
        num_sites (int):
            Size of system (usually in number of prims in supercell, but
            can be anything representative, i.e. number of sites)
        sublattices (list of Sublattice)
            Sublattices of the ensemble sampled.
        natural_parameters (ndarray):
                Array of natural parameters used in the ensemble.
        metadata (dict):
            Dictionary of metadata from the MC run that generated the samples.
    """

    def __init__(self, ensemble, sample_trace, sampling_metadata=None):
        """Initialize a sample container.

        Args:
            ensemble (Ensemble):
                Ensemble object used to generate the samples.
            sample_trace (Trace):
                a trace object for the traced values during MC sampling
            sampling_metadata (Ensemble):
                sampling metadata (i.e. ensemble name, mckernel type, etc)
        """
        self._total_steps = 0

        if sampling_metadata is not None:
            if isinstance(sampling_metadata, Metadata):
                self.metadata = sampling_metadata
            else:
                self.metadata = Metadata(SampleContainer.__name__, **sampling_metadata)
        else:
            self.metadata = Metadata(SampleContainer.__name__)

        # TODO these private attributes should be removed once the older format of
        #  sample containers (those that do not save ensembles) is no longer supported
        self._sublattices = None
        self._natural_parameters = None
        self._num_energy_coefs = (
            ensemble.num_energy_coefs if ensemble is not None else None
        )

        self._ensemble = ensemble
        # need counter because can't use shape of arrays when allocating space
        self._nsamples = 0
        self._trace = sample_trace
        self._aux_checkpoint = None
        self._backend = None  # for streaming

    @property
    def sublattices(self):
        """Return the sublattices."""
        if self._ensemble is None:
            return self._sublattices
        return self._ensemble.sublattices

    @property
    def natural_parameters(self):
        """Return the natural parameters."""
        if self._ensemble is None:
            return self._natural_parameters
        return self._ensemble.natural_parameters

    @property
    def ensemble(self):
        """Return the ensemble name."""
        return self._ensemble

    @property
    def num_samples(self):
        """Get the total number of samples."""
        return self._nsamples

    @property
    def total_mc_steps(self):
        """Return the total number of MC steps taken during sampling."""
        return self._total_steps

    @property
    def shape(self):
        """Get the shape of the samples in chain."""
        return self._trace.occupancy.shape[1:]

    @property
    def traced_values(self):
        """Get the names of traced values being sampled."""
        return self._trace.names

    def sampling_efficiency(self, discard=0, flat=True):
        """Return the sampling efficiency for chains.

        If the sampling is thinned by > 1, this value becomes only an estimate for the
        true sampling efficiency, as we do not know the efficiency of the discarded
        samples.
        """
        total_accepted = self._trace.accepted[discard:].sum(axis=0)
        efficiency = total_accepted / (self._nsamples - discard)
        if flat:
            efficiency = efficiency.mean()
        return efficiency

    def get_sampled_structures(self, indices, flat=True):
        """Get sampled structures for MC steps given by indices.

        Args:
            indices (list of int or int):
                A single index or list of indices to obtain sampled structures.
            flat (bool): optional
                If true will flatten chains, and the indices correspond to flattened
                values. If false chain is not flattened, and if multiple walkers where
                used returns a list of list where each inner list has the sampled
                structure for each walker. Default is set to flat True.

        Returns:
            list of Structure or list of list of Structure
        """
        if self.ensemble is None:
            raise RuntimeError(
                "Cannot get structures for SampleContainer with older format!"
            )

        indices = [indices] if isinstance(indices, int) else indices
        occupancies = self.get_occupancies(flat=flat)[indices]
        if flat:
            structures = [
                self.ensemble.processor.structure_from_occupancy(occu)
                for occu in occupancies
            ]
        else:
            structures = [
                [
                    self.ensemble.processor.structure_from_occupancy(occu)
                    for occu in occus
                ]
                for occus in occupancies
            ]
        return structures

    def get_trace_value(self, name, discard=0, thin_by=1, flat=True):
        """Get sampled values of a traced value given by name."""
        value = getattr(self._trace, name)[discard + thin_by - 1 :: thin_by]
        if flat:
            value = self._flatten(value)
        return value

    def mean_trace_value(self, name, discard=0, thin_by=1, flat=True):
        """Get mean of a traced value given by name."""
        return self.get_trace_value(name, discard, thin_by, flat).mean(axis=0)

    def trace_value_variance(self, name, discard=0, thin_by=1, flat=True):
        """Get variance of a traced value given by name."""
        return self.get_trace_value(name, discard, thin_by, flat).var(axis=0)

    def get_occupancies(self, discard=0, thin_by=1, flat=True):
        """Get the occupancy chain of the samples."""
        return self.get_trace_value("occupancy", discard, thin_by, flat)

    def get_enthalpies(self, discard=0, thin_by=1, flat=True):
        """Get the generalized enthalpy changes from samples in chain."""
        return self.get_trace_value("enthalpy", discard, thin_by, flat)

    def get_feature_vectors(self, discard=0, thin_by=1, flat=True):
        """Get the feature vector changes from samples in chain."""
        return self.get_trace_value("features", discard, thin_by, flat)

    def get_energies(self, discard=0, thin_by=1, flat=True):
        """Get the energies from samples in chain."""
        if len(self.natural_parameters) == self._num_energy_coefs:
            return self.get_enthalpies(discard, thin_by, flat)
        # otherwise we have to calculate without the additional terms...
        feature_trace = self.get_feature_vectors(discard, thin_by, flat=False)
        energies = np.expand_dims(
            np.vstack(
                [
                    np.tensordot(
                        features[:, : self._num_energy_coefs],
                        self.natural_parameters[: self._num_energy_coefs],
                        axes=([1], [0]),
                    )
                    for features in feature_trace
                ]
            ),
            axis=-1,
        )
        if flat:
            energies = self._flatten(energies)
        return energies

    def get_sublattice_compositions(self, sublattice, discard=0, thin_by=1, flat=True):
        """Get the compositions of a specific sublattice."""
        counts = self.get_sublattice_species_counts(sublattice, discard, thin_by, flat)
        return counts / len(sublattice.sites)

    def get_compositions(self, discard=0, thin_by=1, flat=True):
        """Get the compositions for each occupancy in the chain."""
        counts = self.get_species_counts(discard, thin_by, flat)
        return {spec: count / self.shape[1] for spec, count in counts.items()}

    def mean_enthalpy(self, discard=0, thin_by=1, flat=True):
        """Get the mean generalized enthalpy."""
        return self.get_enthalpies(discard, thin_by, flat).mean(axis=0)

    def enthalpy_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance in enthalpy."""
        return self.get_enthalpies(discard, thin_by, flat).var(axis=0)

    def mean_energy(self, discard=0, thin_by=1, flat=True):
        """Calculate the mean energy from samples."""
        return self.get_energies(discard, thin_by, flat).mean(axis=0)

    def energy_variance(self, discard=0, thin_by=1, flat=True):
        """Calculate the variance of sampled energies."""
        return self.get_energies(discard, thin_by, flat).var(axis=0)

    def mean_feature_vector(self, discard=0, thin_by=1, flat=True):
        """Get the mean feature vector from samples."""
        return self.get_feature_vectors(discard, thin_by, flat).mean(axis=0)

    def feature_vector_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance of feature vector elements."""
        return self.get_feature_vectors(discard, thin_by, flat).var(axis=0)

    def get_orbit_factors(self, function_orbit_ids, discard=0, thin_by=1, flat=True):
        """Get the orbit factor vectors for samples."""
        vals = self.natural_parameters * self.get_feature_vectors(
            discard=discard, thin_by=thin_by, flat=flat
        )
        orbit_factors = np.array(
            [
                np.sum(vals[function_orbit_ids == i])
                for i in range(len(self.natural_parameters))
            ]
        )
        return orbit_factors

    def mean_composition(self, discard=0, thin_by=1, flat=True):
        """Get mean composition for all species regardless of sublattice."""
        comps = self.get_compositions(discard, thin_by, flat)
        return {spec: comp.mean(axis=0) for spec, comp in comps.items()}

    def composition_variance(self, discard=0, thin_by=1, flat=True):
        """Get the variance in composition of all species."""
        comps = self.get_compositions(discard, thin_by, flat)
        return {spec: comp.var(axis=0) for spec, comp in comps.items()}

    def mean_sublattice_composition(self, sublattice, discard=0, thin_by=1, flat=True):
        """Get the mean composition of a specific sublattice."""
        return self.get_sublattice_compositions(
            sublattice, discard, thin_by, flat
        ).mean(axis=0)

    def sublattice_composition_variance(
        self, sublattice, discard=0, thin_by=1, flat=True
    ):
        """Get the variance in composition of a specific sublattice."""
        return self.get_sublattice_compositions(sublattice, discard, thin_by, flat).var(
            axis=0
        )

    def get_minimum_enthalpy(self, discard=0, thin_by=1, flat=True):
        """Get the minimum energy from samples."""
        return self.get_enthalpies(discard, thin_by, flat).min(axis=0)

    def get_minimum_enthalpy_occupancy(self, discard=0, thin_by=1, flat=True):
        """Find the occupancy with minimum energy from samples."""
        inds = self.get_enthalpies(discard, thin_by, flat).argmin(axis=0)
        if flat:
            occus = self.get_occupancies(discard, thin_by, flat)[inds]
        else:
            occus = self.get_occupancies(discard, thin_by, flat)[
                inds, np.arange(self.shape[0])
            ][0]
        return occus

    def get_minimum_energy(self, discard=0, thin_by=1, flat=True):
        """Get the minimum energy from samples."""
        return self.get_energies(discard, thin_by, flat).min(axis=0)

    def get_minimum_energy_occupancy(self, discard=0, thin_by=1, flat=True):
        """Find the occupancy with minimum energy from samples."""
        inds = self.get_energies(discard, thin_by, flat).argmin(axis=0)
        if flat:
            occus = self.get_occupancies(discard, thin_by, flat)[inds]
        else:
            occus = self.get_occupancies(discard, thin_by, flat)[
                inds, np.arange(self.shape[0])
            ][0]
        return occus

    def get_species_counts(self, discard=0, thin_by=1, flat=True):
        """Get the species counts for each occupancy in the chain."""
        samples = (self.num_samples - discard) // thin_by
        shape = self.shape[0] * samples if flat else (self.shape[0], samples)
        counts = defaultdict(lambda: np.zeros(shape=shape))
        for sublattice in self.sublattices:
            subcounts = self.get_sublattice_species_counts(
                sublattice, discard, thin_by, flat
            )
            for species, count in zip(sublattice.species, subcounts.T):
                counts[species] += count
        return counts

    def get_sublattice_species_counts(
        self, sublattice, discard=0, thin_by=1, flat=True
    ):
        """Get the counts of each species in a sublattices.

        Returns:
            ndarray: where last axis is the count for each species in the
            same order as the underlying site space.
        """
        if sublattice not in self.sublattices:
            raise ValueError(
                "Sublattice provided is not recognized.\n Provide one included"
                " in the sublattices attribute of this SampleContainer."
            )
        occus = self.get_occupancies(discard, thin_by, flat=False)
        counts = np.zeros((*occus.shape[:-1], len(sublattice.site_space)))
        #  This can probably be re-written in a clean/faster way
        for i, occupancies in enumerate(occus):
            for j, occupancy in enumerate(occupancies):
                codes, count = np.unique(
                    occupancy[sublattice.sites], return_counts=True
                )
                # check for zero counts
                if len(codes) != len(sublattice.site_space):
                    missed = list(set(sublattice.encoding) - set(codes))
                    codes = np.append(codes, missed)
                    count = np.append(count, len(missed) * [0])

                original_codes = sublattice.encoding.tolist()
                order = [codes.tolist().index(code) for code in original_codes]
                counts[i][j] = count[order]  # order them accordingly
        if flat:
            counts = self._flatten(counts)
        return counts

    def save_sampled_trace(self, trace, thinned_by):
        """Save a sampled trace.

        Args:
            trace (Trace)
                Trace of sampled values
            thinned_by (int):
                the amount that the sampling was thinned by. Used to update
                the total mc iterations.
        """
        for name, value in trace.items():
            getattr(self._trace, name)[self._nsamples] = value
        self._nsamples += 1
        self._total_steps += thinned_by

    def clear(self):
        """Clear all samples from container."""
        self._total_steps = 0
        self._total_steps = 0
        self._nsamples = 0
        for name, value in self._trace.items():
            setattr(
                self._trace, name, np.empty((0, *value.shape[1:]), dtype=value.dtype)
            )

    def allocate(self, nsamples):
        """Allocate more space in arrays for more samples."""
        for name, value in self._trace.items():
            arr = np.empty((nsamples, *value.shape[1:]), dtype=value.dtype)
            setattr(self._trace, name, np.append(value, arr, axis=0))

    def vacuum(self):
        """Remove any trailing allocated space that has not been used."""
        for name, value in self._trace.items():
            setattr(self._trace, name, value[: self._nsamples])

    def flush_to_backend(self, backend):
        """Flush current samples and trace to backend file.

        Args:
            backend (object):
                backend file object, currently only hdf5 supported.
        """
        start = backend["trace"].attrs["nsamples"]
        end = self._nsamples + start
        for name, value in self._trace.items():
            backend["trace"][name][start:end] = value

        backend["trace"].attrs["total_mc_steps"] += self._total_steps
        backend["trace"].attrs["nsamples"] += self._nsamples
        backend.flush()

        self._total_steps = 0
        self._nsamples = 0

    def get_backend(self, file_path, alloc_nsamples=0, swmr_mode=False):
        """Get a backend file object.

        Currently only hdf5 files supported

        Args:
            file_path (str):
                path to backend file.
            alloc_nsamples (int): optional
                number of new samples to allocate. Will only extend datasets
                if number given is larger than space left to write samples
                into.
            swmr_mode (bool): optional
                If true allows to read file from other processes. Single Writer
                Multiple Readers.

        Returns:
            h5.File object
        """
        if os.path.isfile(file_path):
            backend = self._check_backend(file_path)
            trace_grp = backend["trace"]
            available = len(trace_grp["occupancy"]) - trace_grp.attrs["nsamples"]
            # this probably fails since maxshape is not set.
            if available < alloc_nsamples:
                SampleContainer._grow_backend(backend, alloc_nsamples - available)
        else:
            backend = h5py.File(file_path, "w-", libver="latest")
            self._init_backend(backend, alloc_nsamples)

        if swmr_mode:
            backend.swmr_mode = swmr_mode
        return backend

    def _check_backend(self, file_path):
        """Check if existing backend file is populated correctly."""
        backend = h5py.File(file_path, mode="r+", libver="latest")
        if self.shape != backend["trace"]["occupancy"].shape[1:]:
            shape = backend["trace"]["occupancy"].shape[1:]
            backend.close()
            raise RuntimeError(
                f"Backend file {file_path} has incompatible dimensions "
                f"{self.shape}, {shape}."
            )
        return backend

    def _init_backend(self, backend, nsamples):
        """Initialize a backend file."""
        metadata = backend.create_group("metadata")
        metadata.create_dataset(
            "ensemble", data=json.dumps(self.ensemble, cls=MontyEncoder)
        )
        metadata.create_dataset(
            "sampling_metadata", data=json.dumps(self.metadata, cls=MontyEncoder)
        )
        trace_grp = backend.create_group("trace")
        for name, value in self._trace.items():
            trace_grp.create_dataset(
                name,
                shape=(nsamples, *value.shape[1:]),
                dtype=value.dtype,
                maxshape=(None, *value.shape[1:]),
            )
        trace_grp.attrs["nsamples"] = 0
        trace_grp.attrs["total_mc_steps"] = 0
        backend.flush()  # flush initialization

    @staticmethod
    def _grow_backend(backend, nsamples):
        """Extend space available in a backend file."""
        for name in backend["trace"]:
            backend["trace"][name].resize(
                len(backend["trace"][name]) + nsamples, axis=0
            )

    @staticmethod
    def _flatten(traced_values):
        """Flatten values in trace values with multiple walkers."""
        shape_l = list(traced_values.shape[1:])
        shape_l[0] = np.prod(traced_values.shape[:2])
        return np.squeeze(traced_values.reshape(shape_l))

    def __len__(self):
        """Return the number of samples."""
        return self._nsamples

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        container_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "ensemble": self.ensemble.as_dict(),
            "metadata": self.metadata.as_dict(),
            "total_mc_steps": self._total_steps,
            "nsamples": self._nsamples,
            "trace": jsanitize(self._trace.as_dict()),
            "aux_checkpoint": self._aux_checkpoint,
        }
        # TODO need to think how to generally serialize the aux checkpoint
        return container_d

    @classmethod  # TODO remove this when old container format (without ensemble) are no longer supported
    def _from_legacy(
        cls, ensemble, sublattices, trace, metadata, nat_params, n_energy_coefs
    ):
        """Load a legacy container."""
        if ensemble is None:
            warnings.warn(
                "This SampleContainer was created with a previous version of smol.\n"
                "This format is deprecated and will not be supported in the future.\n"
                "To update your SampleContainer, generate it using the from_dict or "
                "from_hdf5 methods passing the Ensemble that was used to generate samples"
                " using the `ensemble` keyword argument, and re-save it.",
                FutureWarning,
            )
            container = cls(None, trace, metadata)
            container._sublattices = sublattices
            container._natural_parameters = nat_params
            container._num_energy_coefs = n_energy_coefs
        else:
            # check that sublattices match (this does not check if active sites match!)
            if not all(sl in sublattices for sl in ensemble.sublattices) or len(
                ensemble.sublattices
            ) != len(sublattices):
                raise ValueError(
                    "Sublattices in Ensemble object passed do not match, the once saved. \n "
                    "Make sure you are passing the correct Ensemble object."
                )
            container = cls(ensemble, trace, metadata)
        return container

    @classmethod
    def from_dict(cls, d, ensemble=None):
        """Instantiate a SampleContainer from dict representation.

        Args:
            d (dict):
                dictionary representation.
            ensemble (Ensemble): optional
                The ensemble object to used for generating samples. Only needed to
                update legacy files.
        Returns:
            SampleContainer
        """
        trace = Trace(**{key: np.array(val) for key, val in d["trace"].items()})
        d["metadata"] = MontyDecoder().process_decoded(d["metadata"])

        if not isinstance(d["metadata"], Metadata):  # backwards compatibility
            d["metadata"] = Metadata(cls.__name__, **d["metadata"])

        if "sublattices" in d.keys():  # legacy file
            sublattices = [
                Sublattice.from_dict(sublatt) for sublatt in d["sublattices"]
            ]
            container = cls._from_legacy(
                ensemble,
                sublattices,
                trace,
                d["metadata"],
                d["natural_parameters"],
                d["num_energy_coefs"],
            )
        else:
            ensemble = Ensemble.from_dict(d["ensemble"])
            container = cls(ensemble, trace, d["metadata"])

        # set the container internals
        container._total_steps = d["total_mc_steps"]
        container._nsamples = trace.occupancy.shape[0]
        container._aux_checkpoint = d["aux_checkpoint"]
        return container

    def to_hdf5(self, file_path):
        """Save SampleContainer as an HDF5 file.

        Args:
            file_path (str):
                path to file save location. If file exists and dimensions
                match samples will be appended.
        """
        # keep these to not reset writing
        total_mc_steps, nsamples = self._total_steps, self._nsamples
        backend = self.get_backend(file_path, self.num_samples)
        self.flush_to_backend(backend)
        self._total_steps, self._nsamples = total_mc_steps, nsamples
        backend.close()

    @classmethod
    def from_hdf5(cls, file_path, swmr_mode=True, ensemble=None):
        """Instantiate a SampleContainer from an hdf5 file.

        Args:
            file_path (str):
                path to file
            swmr_mode (bool): optional
                If true allows to read file from other processes. Single Writer
                Multiple Readers.
            ensemble (Ensemble): optional
                The ensemble object to used for generating samples. Only needed to
                update legacy files.

        Returns:
            SampleContainer
        """
        with h5py.File(file_path, "r", swmr=swmr_mode) as f:
            # Check if written states matches the size of datasets
            nsamples = f["trace"].attrs["nsamples"]
            if len(f["trace"]["occupancy"]) > nsamples:
                warnings.warn(
                    f"The hdf5 file provided appears to be from an unifinished"
                    f" MC run.\n Only {nsamples} of "
                    f" {len(f['trace']['occupancy'])} samples "
                    f"have been written and will be loaded.",
                    UserWarning,
                )
            trace = Trace(
                **{name: value[:nsamples] for name, value in f["trace"].items()}
            )

            if "sublattices" in f.keys():  # legacy file
                sampling_metadata = json.loads(
                    f["sampling_metadata"][()], cls=MontyDecoder
                )
                if not isinstance(
                    sampling_metadata, Metadata
                ):  # backwards compatibility
                    sampling_metadata = Metadata(cls.__name__, **sampling_metadata)

                sublattices = [
                    Sublattice.from_dict(s) for s in json.loads(f["sublattices"][()])
                ]
                container = cls._from_legacy(
                    ensemble,
                    sublattices,
                    trace,
                    sampling_metadata,
                    f["natural_parameters"][()],
                    f["natural_parameters"].attrs["num_energy_coefs"],
                )
            else:
                sampling_metadata = Metadata.from_dict(
                    json.loads(f["metadata"]["sampling_metadata"][()])
                )
                ensemble = Ensemble.from_dict(json.loads(f["metadata"]["ensemble"][()]))
                container = cls(ensemble, trace, sampling_metadata)

            container._nsamples = nsamples
            container._total_steps = f["trace"].attrs["total_mc_steps"]

        return container



"""
Class implementations to create Cluster Expansions.

The cofe (pronounced coffee) package contains all the necessary classes and
functions to define and fit cluster expansions for crystalline materials.
"""

from smol.cofe.expansion import ClusterExpansion, RegressionData
from smol.cofe.space.basis import available_site_basis_sets
from smol.cofe.space.clusterspace import ClusterSubspace, PottsSubspace
from smol.cofe.wrangling.wrangler import StructureWrangler

__all__ = [
    "ClusterSubspace",
    "PottsSubspace",
    "StructureWrangler",
    "ClusterExpansion",
    "RegressionData",
    "available_site_basis_sets",
]



import math
from collections import Counter
from random import choice, choices

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Composition

from smol.capp.generate.random import _gen_neutral_occu, _gen_unconstrained_ordered_occu
from smol.cofe.space.domain import SiteSpace
from smol.moca.kernel import mcusher
from smol.moca.kernel.base import ALL_MCUSHERS
from smol.moca.kernel.bias import SquareChargeBias
from smol.moca.sublattice import Sublattice
from tests.utils import assert_pickles

mcmcusher_classes = [getattr(mcusher, i) for i in ALL_MCUSHERS]
num_sites = 100


@pytest.fixture
def all_sublattices(rng):
    # generate two tests sublattices
    sites = np.arange(num_sites)
    sites1 = rng.choice(sites, size=num_sites // 3)
    sites2 = rng.choice(np.setdiff1d(sites, sites1), size=num_sites // 4)
    sites3 = np.setdiff1d(sites, np.concatenate((sites1, sites2)))
    site_space1 = SiteSpace(Composition({"A": 0.1, "B": 0.4, "C": 0.3, "D": 0.2}))
    site_space2 = SiteSpace(Composition({"A": 0.1, "B": 0.4, "E": 0.5}))
    site_space3 = SiteSpace(Composition({"G": 1}))
    active_sublattices = [
        Sublattice(site_space1, sites1),
        Sublattice(site_space2, sites2),
    ]
    inactive_sublattices = [Sublattice(site_space3, sites3)]
    return active_sublattices, inactive_sublattices


@pytest.fixture
def all_sublattices_lmtpo():  # Do a test on sampling probabilities.
    # generate two tests sublattices
    sites = np.arange(6, dtype=int)
    sites1 = np.random.choice(sites, size=3, replace=False)
    sites2 = np.setdiff1d(sites, sites1)
    site_space1 = SiteSpace(Composition({"Li+": 2 / 3, "Zr4+": 1 / 6, "Mn3+": 1 / 6}))
    site_space2 = SiteSpace(Composition({"O2-": 5 / 6, "F-": 1 / 6}))
    active_sublattices = [
        Sublattice(site_space1, sites1),
        Sublattice(site_space2, sites2),
    ]
    inactive_sublattices = []
    return active_sublattices, inactive_sublattices


@pytest.fixture
def rand_occu(all_sublattices, rng):
    # generate a random occupancy according to the sublattices
    occu = _gen_unconstrained_ordered_occu(
        all_sublattices[0] + all_sublattices[1], rng=rng
    )
    return occu, all_sublattices[1][0].sites  # return indices of fixed sites


@pytest.fixture
def rand_occu_lmtpo(all_sublattices_lmtpo, rng):
    # generate a random occupancy according to the sublattices
    occu = _gen_neutral_occu(
        all_sublattices_lmtpo[0] + all_sublattices_lmtpo[1], rng=rng
    )
    return occu, []  # return indices of fixed sites


@pytest.fixture(params=mcmcusher_classes)
def mcmcusher(request, all_sublattices):
    # instantiate mcmcushers to test
    kwargs = {}
    if request.param == mcusher.TableFlip:
        kwargs["swap_weight"] = 0
        # return request.param(all_sublattices[0] + all_sublattices[1], swap_weight=0)
    elif request.param == mcusher.MultiStep:
        ushers = [
            c
            for c in mcmcusher_classes
            # avoid TableFlip to avoid further checking in differeting proposal probabilities
            if c not in (mcusher.MultiStep, mcusher.Composite, mcusher.TableFlip)
        ]
        kwargs["mcusher"] = choice(ushers)(all_sublattices[0] + all_sublattices[1])
        kwargs["step_lengths"] = np.random.randint(1, 4, size=2)
    elif request.param == mcusher.Composite:
        kwargs["mcushers"] = choices(
            [
                klass(all_sublattices[0] + all_sublattices[1])
                for klass in mcmcusher_classes
                if klass
                not in (mcusher.MultiStep, mcusher.Composite, mcusher.TableFlip)
            ],
            k=np.random.randint(2, len(mcmcusher_classes) - 2),
        )
    return request.param(all_sublattices[0] + all_sublattices[1], **kwargs)


@pytest.fixture
def table_flip(all_sublattices_lmtpo):
    return mcusher.TableFlip(
        all_sublattices_lmtpo[0] + all_sublattices_lmtpo[1],
        optimize_basis=True,
        table_ergodic=True,
        swap_weight=0.2,
    )


def test_bad_propabilities(mcmcusher):
    with pytest.raises(ValueError):
        mcmcusher.sublattice_probabilities = [0.6, 0.1]
    with pytest.raises(AttributeError):
        mcmcusher.sublattice_probabilities = [0.5, 0.2, 0.3]
    with pytest.raises(AttributeError):
        _ = mcusher.Flip(mcmcusher.sublattices, [0.5])
    with pytest.raises(ValueError):
        _ = mcusher.Flip(mcmcusher.sublattices, [0.5, 0.2])


def test_propose_step(mcmcusher, rand_occu):
    occu, fixed_sites = rand_occu
    iterations = 50000
    # test with 50/50 probability
    flipped_sites = []
    count1, count2 = 0, 0
    total = 0
    for i in range(iterations):
        step = mcmcusher.propose_step(occu)
        assert len(step) == len({s for s, c in step})
        # No duplicate site allowed!
        for flip in step:
            assert flip[1] != occu[flip[0]]
            if flip[0] in mcmcusher.active_sublattices[0].active_sites:
                count1 += 1
                assert flip[1] in mcmcusher.active_sublattices[0].encoding
            elif flip[0] in mcmcusher.active_sublattices[1].active_sites:
                count2 += 1
                assert flip[1] in mcmcusher.active_sublattices[1].encoding
            else:
                raise RuntimeError(
                    "Something went wrong in proposing a step site proposed in {step} "
                    "is not in any of the allowed sites"
                )
            total += 1
            flipped_sites.append(flip[0])

    # check probabilities seem sound
    if not isinstance(mcmcusher, mcusher.TableFlip):
        assert count1 / total == pytest.approx(0.5, abs=2e-2)
        assert count2 / total == pytest.approx(0.5, abs=2e-2)
    else:
        # Because Table flip is equal per-direction.
        assert count1 / total == pytest.approx(0.6, abs=2e-2)
        assert count2 / total == pytest.approx(0.4, abs=2e-2)

    # check that every site was flipped at least once
    assert all(
        i in flipped_sites for i in np.setdiff1d(np.arange(num_sites), fixed_sites)
    )

    # make sure fixed sites remain the same
    assert all(i not in fixed_sites for i in flipped_sites)

    # Now check with a sublattice bias
    if isinstance(mcmcusher, mcusher.Composite):
        for usher in mcmcusher.mcushers:
            usher.sublattice_probabilities = [0.8, 0.2]
    else:
        mcmcusher.sublattice_probabilities = [0.8, 0.2]
    flipped_sites = []
    count1, count2 = 0, 0
    total = 0
    for i in range(iterations):
        step = mcmcusher.propose_step(occu)
        for flip in step:
            if flip[0] in mcmcusher.active_sublattices[0].active_sites:
                count1 += 1
                assert flip[1] in mcmcusher.active_sublattices[0].encoding
            elif flip[0] in mcmcusher.active_sublattices[1].sites:
                count2 += 1
                assert flip[1] in mcmcusher.active_sublattices[1].encoding
            else:
                raise RuntimeError(
                    "Something went wrong in proposing"
                    f"a step site proposed in {step} is"
                    " not in any of the allowed sites"
                )
            total += 1
            flipped_sites.append(flip[0])
    if not isinstance(mcmcusher, mcusher.TableFlip):
        assert count1 / total == pytest.approx(0.8, abs=2e-2)
        assert count2 / total == pytest.approx(0.2, abs=2e-2)


def test_table_flip_factors():
    sites1 = np.array([0, 1, 2])
    sites2 = np.array([3, 4, 5])
    site_space1 = SiteSpace(Composition({"Li+": 2 / 3, "Zr4+": 1 / 6, "Mn3+": 1 / 6}))
    site_space2 = SiteSpace(Composition({"O2-": 5 / 6, "F-": 1 / 6}))
    sublattices = [Sublattice(site_space1, sites1), Sublattice(site_space2, sites2)]

    tf = mcusher.TableFlip(sublattices, optimize_basis=True, table_ergodic=True)
    # Case 1:
    occu1 = np.array([0, 0, 1, 0, 0, 0])
    step1 = [(2, 2), (4, 1)]
    assert np.isclose(
        tf.compute_log_priori_factor(occu1, step1), np.log(3 / 2)
    )  # forth p=1/3, back p=1/2
    # Case 2:
    occu2 = np.array([0, 0, 2, 1, 0, 0])
    step2 = [(2, 1), (3, 0)]
    assert np.isclose(
        tf.compute_log_priori_factor(occu2, step2), np.log(2 / 3)
    )  # forth p=1/2, back p=1/3
    # Case 3:
    occu3 = np.array([0, 0, 2, 1, 0, 0])
    step3 = [(2, 0), (4, 1), (5, 1)]
    assert np.isclose(
        tf.compute_log_priori_factor(occu3, step3), np.log(2 / 9)
    )  # forth p=1/2, back p=1/9
    # Case 4:
    occu4 = np.array([0, 0, 0, 1, 1, 1])
    step4 = [(0, 2), (4, 0), (5, 0)]
    assert np.isclose(
        tf.compute_log_priori_factor(occu4, step4), np.log(9 / 2)
    )  # forth p=1/9, back p=1/2
    # Case 5:
    occu5 = np.array([0, 0, 2, 1, 0, 0])
    step5 = [(2, 0), (0, 2)]
    assert np.isclose(tf.compute_log_priori_factor(occu5, step5), 0)  # forth p=back p.


def test_table_flip(table_flip, rand_occu_lmtpo):
    def get_n(occu, sublattices):
        sl1, sl2 = sublattices
        n = np.array(
            [
                (occu[sl1.sites] == 0).sum(),
                (occu[sl1.sites] == 1).sum(),
                (occu[sl1.sites] == 2).sum(),
                (occu[sl2.sites] == 0).sum(),
                (occu[sl2.sites] == 1).sum(),
            ],
            dtype=int,
        )
        return n

    def get_hash(a):
        return tuple(a.tolist())

    def get_n_states(n):
        assert n[:3].sum() == 3
        assert n[3:].sum() == 3
        return math.comb(3, n[0]) * math.comb(3 - n[0], n[1]) * math.comb(3, n[3])

    occu = rand_occu_lmtpo[0].copy()
    bias = SquareChargeBias(table_flip.sublattices)
    o_counter = Counter()
    n_counter = Counter()

    # Uniformly random kernel.
    # print("Sublattices:", table_flip.sublattices)
    # print("flip table:", table_flip.flip_table)
    l = 100000
    for i in range(l):
        assert bias.compute_bias(occu) == 0
        step = table_flip.propose_step(occu)
        n = get_n(occu, table_flip.sublattices)
        # print("occu:", occu)
        # print("n:", n)
        # print("step:", step)
        flip_id, direction = table_flip._get_flip_id(occu, step)
        occu_next = occu.copy()
        for s_id, code in step:
            occu_next[s_id] = code
        # print("occu_next:", occu_next)
        # print("n_next:", get_n(occu_next))
        dn = get_n(occu_next, table_flip.sublattices) - n
        # Check dn is always correct.
        if flip_id == -1:
            assert direction == 0
            npt.assert_array_equal(dn, 0)
            if len(step) == 2:
                assert np.any(n != 6)
                assert np.all(n >= 0)
            else:
                assert len(step) == 0
        else:
            dd = -2 * direction + 1
            npt.assert_array_equal(dd * table_flip.flip_table[flip_id, :], dn)

        n_counter[get_hash(n)] += 1
        o_counter[get_hash(occu)] += 1

        log_priori = table_flip.compute_log_priori_factor(occu, step)
        # Null step might still exist.
        # assert len(step) > 0
        if log_priori >= 0 or log_priori >= np.log(np.random.rand()):
            # Accepted.
            occu = occu_next.copy()

    # When finished, see if distribution is correct.
    assert len(n_counter) == 3
    n_occus = []
    for n_hash in n_counter.keys():
        n = np.array(n_hash, dtype=int)
        n_occus.append(get_n_states(n))
    n_occus = np.array(n_occus)
    assert len(o_counter) == sum(n_occus)
    o_count_av = l / sum(n_occus)
    npt.assert_allclose(np.array(list(o_counter.values())) / o_count_av, 1, atol=0.1)
    n_counts = np.array(list(n_counter.values()))
    r_counts = n_counts / n_counts.sum()
    r_occus = n_occus / n_occus.sum()
    npt.assert_allclose(r_counts, r_occus, atol=0.1)
    # print("r_counts:", r_counts)
    # print("r_occus:", r_occus)
    # print("occupancies:", o_counter)
    # assert False
    # Read numerical values, they are acceptable.


def test_pickles(mcmcusher):
    assert_pickles(mcmcusher)



import numpy as np
import pytest

from smol.utils.cluster.container import (
    FloatArray1DContainer,
    FloatArray2DContainer,
    IntArray1DContainer,
    IntArray2DContainer,
    OrbitContainer,
)


@pytest.mark.parametrize(
    "IntArrayContainer, dim", [(IntArray1DContainer, 1), (IntArray2DContainer, 2)]
)
def test_int_container(IntArrayContainer, dim, rng):
    # Use np.int32 to ensure cross-platform compatibility.
    arrays = tuple(
        rng.integers(1, 5, size=dim * (rng.integers(1, 5),)).astype(np.int32)
        for _ in range(10)
    )
    container = IntArrayContainer(arrays)
    assert len(container) == 10

    # test setting the same sized list (no memory reallocation)
    new_arrays = tuple(
        rng.integers(1, 5, size=dim * (rng.integers(1, 5),)).astype(np.int32)
        for _ in range(10)
    )
    container.set_arrays(new_arrays)

    assert len(container) == 10

    # test setting a different sized list (needs memory reallocation)
    new_arrays = tuple(
        rng.integers(1, 5, size=dim * (rng.integers(1, 5),)).astype(np.int32)
        for _ in range(12)
    )
    container.set_arrays(new_arrays)

    assert len(container) == 12

    with pytest.raises(ValueError):
        new_arrays = tuple(
            rng.integers(1, 5, size=(dim + 1) * (rng.integers(1, 5),)).astype(np.int32)
            for _ in range(12)
        )
        container.set_arrays(new_arrays)


@pytest.mark.parametrize(
    "FloatContainer, dim", [(FloatArray1DContainer, 1), (FloatArray2DContainer, 2)]
)
def test_float_container(FloatContainer, dim, rng):
    arrays = tuple(rng.random(tuple(rng.integers(1, 5, size=dim))) for _ in range(10))
    container = FloatContainer(arrays)
    assert len(container) == 10

    # test setting the same sized list (no memory reallocation)
    new_arrays = tuple(
        rng.random(tuple(rng.integers(1, 5, size=dim))) for _ in range(10)
    )
    container.set_arrays(new_arrays)

    assert len(container) == 10

    # test setting a different sized list (needs memory reallocation)
    new_arrays = tuple(
        rng.random(tuple(rng.integers(1, 5, size=dim))) for _ in range(12)
    )
    container.set_arrays(new_arrays)
    assert len(container) == 12

    with pytest.raises(ValueError):
        array_list = tuple(
            rng.random(tuple(rng.integers(1, 5, size=dim + 1))) for _ in range(10)
        )
        container.set_arrays(array_list)


def test_orbit_container(rng):
    orbit_data = []
    for i in range(10):
        orbit_id = i
        bit_id = rng.integers(1, 5, dtype=int)
        correlation_tensors = rng.random(tuple(rng.integers(1, 5, size=2))).astype(
            np.float64
        )
        tensor_indices = rng.integers(1, 5, size=correlation_tensors.shape[0]).astype(
            np.int32
        )
        orbit_data.append((orbit_id, bit_id, correlation_tensors, tensor_indices))

    container = OrbitContainer(tuple(orbit_data))
    assert len(container) == 10

    # test setting the same sized list (no memory reallocation)
    new_orbit_data = []
    for i in range(10):
        orbit_id = i
        bit_id = rng.integers(1, 5, dtype=int)
        correlation_tensors = rng.random(tuple(rng.integers(1, 5, size=2))).astype(
            np.float64
        )
        tensor_indices = rng.integers(1, 5, size=correlation_tensors.shape[0]).astype(
            np.int32
        )
        new_orbit_data.append((orbit_id, bit_id, correlation_tensors, tensor_indices))

    container.set_orbits(tuple(new_orbit_data))
    assert len(container) == 10

    # test setting a different sized list (needs memory reallocation)
    new_orbit_data = []
    for i in range(12):
        orbit_id = i
        bit_id = rng.integers(1, 5, dtype=int)
        correlation_tensors = rng.random(tuple(rng.integers(1, 5, size=2))).astype(
            np.float64
        )
        tensor_indices = rng.integers(1, 5, size=correlation_tensors.shape[0]).astype(
            np.int32
        )
        new_orbit_data.append((orbit_id, bit_id, correlation_tensors, tensor_indices))

    container.set_orbits(tuple(new_orbit_data))
    assert len(container) == 12

    with pytest.raises(TypeError):
        new_orbit_data[-1] = (
            "S",
            bit_id,
            correlation_tensors,
            tensor_indices,
        )  # orbit_id must be an int
        container.set_orbits(tuple(new_orbit_data))

    with pytest.raises(TypeError):
        new_orbit_data[-1] = (
            orbit_id,
            "X",
            correlation_tensors,
            tensor_indices,
        )  # bit_id must be an int
        container.set_orbits(tuple(new_orbit_data))

    with pytest.raises(TypeError):
        new_orbit_data[-1] = (orbit_id, bit_id, "X", tensor_indices)
        container.set_orbits(
            tuple(new_orbit_data)
        )  # correlation_tensors must be a numpy array

    with pytest.raises(TypeError):
        new_orbit_data[-1] = (orbit_id, bit_id, correlation_tensors, "X")
        container.set_orbits(
            tuple(new_orbit_data)
        )  # tensor_indices must be a numpy array

    with pytest.raises(ValueError):
        new_orbit_data[-1] = (
            orbit_id,
            bit_id,
            rng.integers(1, 5, size=10),
            tensor_indices,
        )
        container.set_orbits(tuple(new_orbit_data))  # correlation_tensors must be 2D

    with pytest.raises(ValueError):
        new_orbit_data[-1] = (
            orbit_id,
            bit_id,
            correlation_tensors,
            rng.integers(1, 5, size=(2, 2)),
        )
        container.set_orbits(tuple(new_orbit_data))  # tensor_indices must be 1D



"""Test objective functions generation."""

import numpy as np
import numpy.testing as npt

from smol.capp.generate.groundstate.upper_bound.objectives import (
    get_terms_from_chemical_potentials,
    get_terms_from_decomposition_processor,
    get_terms_from_ewald_processor,
    get_terms_from_expansion_processor,
)
from smol.capp.generate.groundstate.upper_bound.terms import (
    get_auxiliary_variable_values,
    get_expression_and_auxiliary_from_terms,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_occupancy_from_variables,
    get_variables_from_sublattices,
)
from smol.moca.processor import ClusterExpansionProcessor

from .utils import (
    evaluate_correlations_from_variable_values,
    get_random_variable_values,
    validate_correlations_from_occupancy,
    validate_interactions_from_occupancy,
)


def test_expansion_upper(solver_test_ensemble, solver_test_initial_occupancy):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    proc = solver_test_ensemble.processor.processors[0]
    if isinstance(proc, ClusterExpansionProcessor):
        terms = get_terms_from_expansion_processor(
            variable_indices,
            expansion_processor=proc,
        )
        grouped_terms = get_terms_from_expansion_processor(
            variable_indices,
            expansion_processor=proc,
            group_output_by_function=True,
        )
    else:
        terms = get_terms_from_decomposition_processor(
            variable_indices,
            decomposition_processor=proc,
        )
        grouped_terms = get_terms_from_decomposition_processor(
            variable_indices,
            decomposition_processor=proc,
            group_output_by_orbit=True,
        )
    objective, aux, aux_indices, aux_cons = get_expression_and_auxiliary_from_terms(
        terms, variables
    )
    for _ in range(20):
        rand_val = get_random_variable_values(solver_test_ensemble.sublattices)
        aux_val = get_auxiliary_variable_values(rand_val, aux_indices)
        rand_occu = get_occupancy_from_variables(
            solver_test_ensemble.sublattices,
            rand_val,
            variable_indices,
        )
        if isinstance(proc, ClusterExpansionProcessor):
            validate_correlations_from_occupancy(proc, rand_occu)
        else:
            validate_interactions_from_occupancy(proc, rand_occu)

        corr = evaluate_correlations_from_variable_values(grouped_terms, rand_val)
        npt.assert_array_almost_equal(
            corr, proc.compute_feature_vector(rand_occu) / proc.size
        )
        variables.value = rand_val
        aux.value = aux_val
        energy_obj = objective.value
        energy_proc = np.dot(proc.compute_feature_vector(rand_occu), proc.coefs)
        assert np.isclose(energy_obj, energy_proc)
        for con in aux_cons:
            assert con.value()


def test_ewald_upper(solver_test_ensemble, solver_test_initial_occupancy):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    proc = solver_test_ensemble.processor.processors[1]
    terms = get_terms_from_ewald_processor(
        variable_indices,
        ewald_processor=proc,
    )
    objective, aux, aux_indices, aux_cons = get_expression_and_auxiliary_from_terms(
        terms, variables
    )
    for inds in aux_indices:
        assert len(inds) == 2  # No more than pair terms.

    for _ in range(20):
        # Should have the same ewald for either neutral or not neutral.
        rand_val = get_random_variable_values(solver_test_ensemble.sublattices)
        aux_val = get_auxiliary_variable_values(rand_val, aux_indices)
        rand_occu = get_occupancy_from_variables(
            solver_test_ensemble.sublattices,
            rand_val,
            variable_indices,
        )
        variables.value = rand_val
        aux.value = aux_val
        energy_obj = objective.value
        energy_proc = proc.compute_feature_vector(rand_occu) * proc.coefs
        assert np.isclose(energy_obj, energy_proc)
        for con in aux_cons:
            assert con.value()


def test_chemical_potentials_upper(solver_test_ensemble, solver_test_initial_occupancy):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    # Only test semi-grand ensemble.
    if solver_test_ensemble.chemical_potentials is not None:
        terms = get_terms_from_chemical_potentials(
            variable_indices,
            solver_test_ensemble._chemical_potentials["table"],
        )
        objective, aux, aux_indices, aux_cons = get_expression_and_auxiliary_from_terms(
            terms, variables
        )
        # No linearize to point terms.
        assert aux is None
        assert len(aux_indices) == 0
        assert len(aux_cons) == 0
        for _ in range(20):
            rand_val = get_random_variable_values(solver_test_ensemble.sublattices)
            rand_occu = get_occupancy_from_variables(
                solver_test_ensemble.sublattices,
                rand_val,
                variable_indices,
            )
            variables.value = rand_val
            energy_obj = objective.value
            # The last number is -1 * chemical work.
            energy_true = (
                -1 * solver_test_ensemble.compute_feature_vector(rand_occu)[-1]
            )
            assert np.isclose(energy_obj, energy_true)



"""Test all bias terms."""

from copy import deepcopy

import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate.random import _gen_unconstrained_ordered_occu
from smol.moca.composition.space import get_oxi_state
from smol.moca.kernel.bias import (
    FugacityBias,
    SquareChargeBias,
    SquareHyperplaneBias,
    mcbias_factory,
)
from tests.utils import assert_pickles

bias_classes = [FugacityBias, SquareChargeBias, SquareHyperplaneBias]


@pytest.fixture(scope="module")
def all_sublattices(ce_processor):
    return ce_processor.get_sublattices()


@pytest.fixture(params=bias_classes)
def mcbias(all_sublattices, request):
    if request.param == SquareHyperplaneBias:
        n_dims = sum(len(sublatt.species) for sublatt in all_sublattices)
        n_cons = max(n_dims - 1, 1)
        a = np.random.randint(low=-10, high=10, size=(n_cons, n_dims))
        b = np.random.randint(low=-10, high=10, size=n_cons)
        return request.param(all_sublattices, a, b)
    return request.param(all_sublattices)


def test_compute_bias_change(mcbias, rng):
    step = []
    occu = _gen_unconstrained_ordered_occu(mcbias.sublattices, rng=rng)
    new_occu = occu.copy()
    rng = np.random.default_rng()
    for _ in range(50):
        s = rng.choice(list(range(len(mcbias.active_sublattices))))
        i = rng.choice(mcbias.active_sublattices[s].sites)
        sp = rng.choice(list(range(len(mcbias.active_sublattices[s].species))))
        step.append((i, sp))
        new_occu[i] = sp
    assert mcbias.compute_bias_change(occu, step) == pytest.approx(
        mcbias.compute_bias(new_occu) - mcbias.compute_bias(occu)
    )


def test_mcbias_factory(all_sublattices):
    for bias in bias_classes:
        if bias == SquareHyperplaneBias:
            n_dims = sum(len(sublatt.species) for sublatt in all_sublattices)
            n_cons = max(n_dims - 1, 1)
            a = np.random.randint(low=-10, high=10, size=(n_cons, n_dims))
            b = np.random.randint(low=-10, high=10, size=n_cons)
            kwargs = {"hyperplane_normals": a, "hyperplane_intercepts": b}
        else:
            kwargs = {}
        assert isinstance(
            mcbias_factory(bias.__name__, all_sublattices, **kwargs), bias
        )


# Tests for FugacityBias
@pytest.fixture(scope="module")
def fugacity_bias(all_sublattices):
    return FugacityBias(all_sublattices)


def test_bad_fugacity_fractions(fugacity_bias):
    fug_fracs = deepcopy(fugacity_bias.fugacity_fractions)
    with pytest.raises(ValueError):
        fug_fracs[0] = {s: v for s, v in list(fug_fracs[0].items())[:-1]}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0] = {sp: 1.1 for sp in fug_fracs[0].keys()}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0] = {"A": 0.5, "D": 0.6}
        fugacity_bias.fugacity_fractions = fug_fracs
    with pytest.raises(ValueError):
        fug_fracs[0]["foo"] = 0.4
        FugacityBias(
            fugacity_bias.sublattices,
            fugacity_fractions=fug_fracs,
        )
    with pytest.raises(ValueError):
        del fug_fracs[0]["foo"]
        FugacityBias(
            fugacity_bias.sublattices,
            fugacity_fractions=fug_fracs,
        )
    with pytest.raises(ValueError):
        fug_fracs[0][str(list(fug_fracs[0].keys())[0])] = 0.0
        fugacity_bias.fugacity_fractions = fug_fracs


def test_build_fu_table(fugacity_bias):
    table = fugacity_bias._build_fu_table(fugacity_bias.fugacity_fractions)
    for sublatt in fugacity_bias.active_sublattices:
        for fus in fugacity_bias.fugacity_fractions:
            if list(sublatt.site_space.keys()) == list(fus.keys()):
                fugacity_fractions = fus
        for i in sublatt.sites:
            for j, species in zip(sublatt.encoding, sublatt.site_space):
                assert fugacity_fractions[species] == table[i, j]


@pytest.fixture(scope="module")
def square_charge_bias(all_sublattices):
    return SquareChargeBias(all_sublattices)


def test_charge_bias(square_charge_bias, rng):
    table = square_charge_bias._c_table
    n_species = max(max(s.encoding) for s in square_charge_bias.sublattices) + 1
    n_sites = sum(len(s.sites) for s in square_charge_bias.sublattices)
    assert table.shape == (n_sites, n_species)
    # All sites on all sublattices must be included in table
    for sublatt in square_charge_bias.sublattices:
        charges = np.array([get_oxi_state(sp) for sp in sublatt.species])
        npt.assert_array_equal(
            table[sublatt.sites[:, None], sublatt.encoding] - charges[None, :], 0
        )
    # Bias should be implemented as negative.
    for _ in range(100):
        occu = _gen_unconstrained_ordered_occu(square_charge_bias.sublattices, rng=rng)
        assert square_charge_bias.compute_bias(occu) <= 1e-6


@pytest.fixture(scope="module")
def square_comp_bias(all_sublattices):
    n_dims = sum(len(sublatt.species) for sublatt in all_sublattices)
    n_cons = max(n_dims - 1, 1)
    a = np.random.randint(low=-10, high=10, size=(n_cons, n_dims))
    b = np.random.randint(low=-10, high=10, size=n_cons)
    return SquareHyperplaneBias(all_sublattices, a, b)


def test_comp_bias(square_comp_bias, rng):
    # Bias should be implemented as negative.
    for _ in range(100):
        occu = _gen_unconstrained_ordered_occu(square_comp_bias.sublattices, rng=rng)
        assert square_comp_bias.compute_bias(occu) <= 1e-6


def test_pickles(mcbias):
    assert_pickles(mcbias)



"""Tools for exhaustive enumeration of symmetrically distinct supercell matrices."""

__author__ = "Luis Barroso-Luque"


import numpy as np

from smol.utils.math import yield_hermite_normal_forms


def enumerate_supercell_matrices(size, symmops, tol=1e-5):
    """Generate all symmetrically distinct supercell matrices of a given size.

    Matrices are given in Hermite normal form following the following work:

    * https://link.aps.org/doi/10.1103/PhysRevB.77.224115
    * https://link.aps.org/doi/10.1103/PhysRevB.80.014120

    Args:
        size (int):
            size of the supercell in multiples of the primitive cell
        symmops (list of SymmOps):
            symmetry operations
        tol (float): optional
            tolerance for checking if a matrix is symmetrically distinct
    Returns:
        list of ndarray: list of supercell matrices with given size
    """
    supercell_matrices = []
    for hnf in yield_hermite_normal_forms(size):
        for symop in symmops:
            hnf_rot = np.linalg.inv(hnf) @ symop.rotation_matrix
            for scm in supercell_matrices:
                unimod = hnf_rot @ scm.T
                if (abs(unimod - np.round(unimod)) < tol).all():
                    break
            else:
                continue
            break
        else:
            supercell_matrices.append(hnf.T)  # supercells in pmg are transpose of hnf

    return supercell_matrices
